"""
Created on 24 Oct 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A tool to specify time intervals that are aligned to the clock - a little bit like crontab.

Checkpoints are specified in the form HH:MM:SS, in a format similar to that for crontab:

** - all values
NN - exactly matching NN
/NN - repeated every NN

For example, **:/5:30 is used to indicate 30 seconds past the minute, every 5 minutes, during every hour.
"""

from datetime import datetime, timedelta

from scs_core.data.datetime import LocalizedDatetime


# --------------------------------------------------------------------------------------------------------------------

class CheckpointGenerator(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, specification):
        pieces = specification.split(':')

        if len(pieces) != 3:
            raise ValueError(specification)

        try:
            hour = CheckpointField.construct(pieces[0], 24)
            minute = CheckpointField.construct(pieces[1], 60)
            second = CheckpointField.construct(pieces[2], 60)

        except ValueError:
            raise ValueError(specification)

        return CheckpointGenerator(hour, minute, second)


    @classmethod
    def is_valid(cls, specification):
        try:
            cls.construct(specification)
            return True

        except (AttributeError, ValueError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, hour, minute, second):
        """
        Constructor
        """
        self.__hour = hour                    # CheckpointField
        self.__minute = minute                # CheckpointField
        self.__second = second                # CheckpointField


    # ----------------------------------------------------------------------------------------------------------------

    def min_interval(self):
        interval = self.__second.interval(1)
        if interval is not None:
            return interval

        interval = self.__minute.interval(60)
        if interval is not None:
            return interval

        interval = self.__hour.interval(3600)
        if interval is not None:
            return interval

        return 86400            # 24 hours


    # ----------------------------------------------------------------------------------------------------------------

    def enclosing_localised_datetime(self, localised_datetime):
        if self.aligns(localised_datetime):
            return localised_datetime

        return self.next_localised_datetime(localised_datetime)


    # if localised_datetime is on a boundary then next_localised_datetime(..) returns a future checkpoint

    def next_localised_datetime(self, localised_datetime):
        # parse...
        date_time = localised_datetime.datetime
        tzinfo = localised_datetime.tzinfo

        date = date_time.date()
        time = date_time.time()

        # compute...
        day_increment, next_hour, next_minute, next_second = self.next(time.hour, time.minute, time.second)

        # construct...
        checkpoint = datetime(date.year, month=date.month, day=date.day,
                              hour=next_hour, minute=next_minute, second=next_second, tzinfo=tzinfo)

        if day_increment:
            checkpoint += timedelta(days=1)

        return LocalizedDatetime(checkpoint)


    # if localised_datetime is on a boundary then prev_localised_datetime(..) returns the same checkpoint

    def prev_localised_datetime(self, localised_datetime):
        # parse...
        date_time = localised_datetime.datetime
        tzinfo = localised_datetime.tzinfo

        date = date_time.date()
        time = date_time.time()

        # compute...
        day_decrement, prev_hour, prev_minute, prev_second = self.prev(time.hour, time.minute, time.second)

        # construct...
        checkpoint = datetime(date.year, month=date.month, day=date.day,
                              hour=prev_hour, minute=prev_minute, second=prev_second, tzinfo=tzinfo)

        if day_decrement:
            checkpoint -= timedelta(days=1)

        return LocalizedDatetime(checkpoint)


    # ----------------------------------------------------------------------------------------------------------------

    def aligns(self, localised_datetime):
        date_time = localised_datetime.datetime
        t = date_time.time()

        return self.__hour.aligns(t.hour) and self.__minute.aligns(t.minute) and self.__second.aligns(t.second) \
            and t.microsecond == 0


    # if hour, minute, second are on a boundary then next(..) returns a future checkpoint

    def next(self, hour, minute, second):
        # seconds...
        next_second = self.__second.next(second)

        # minutes...
        if next_second > second:
            minute -= 1                        # current minute should be tested

        next_minute = self.__minute.next(minute)

        # hours...
        if next_minute > minute:
            hour -= 1                          # current hour should be tested

        next_hour = self.__hour.next(hour)

        # dateline...
        day_increment = self.is_24h() or (next_hour < hour)

        return day_increment, next_hour, next_minute, next_second


    # if hour, minute, second are on a boundary then prev(..) returns the same checkpoint

    def prev(self, hour, minute, second):
        # seconds...
        prev_second = self.__second.prev(second)

        # minutes...
        prev_minute = self.__minute.prev(minute)

        # hours...
        prev_hour = self.__hour.prev(hour)

        # dateline...
        day_decrement = prev_hour > hour

        return day_decrement, prev_hour, prev_minute, prev_second


    # ----------------------------------------------------------------------------------------------------------------

    def is_24h(self):
        return len(self.__hour) == 1 and len(self.__minute) == 1 and len(self.__second) == 1


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CheckpointGenerator:{hour:%s, minute:%s, second:%s}" % (self.__hour, self.__minute, self.__second)


# --------------------------------------------------------------------------------------------------------------------

class CheckpointField(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, specification, size):
        if not 0 < len(specification) < 4:
            raise ValueError(specification)

        # all ticks...
        if specification == "**":
            return CheckpointField([tick for tick in range(0, size, 1)])

        # stepped ticks...
        if specification[0] == '/':
            step = int(specification[1:])

            if not 0 < step < size:
                raise ValueError(specification)

            return CheckpointField([tick for tick in range(0, size, step)])

        # one tick...
        tick = int(specification)

        if not 0 <= tick < size:
            raise ValueError(specification)

        return CheckpointField([tick])


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, ticks):
        """
        Constructor
        """
        self.__ticks = ticks                    # array of int


    def __len__(self):
        return len(self.__ticks)


    # ----------------------------------------------------------------------------------------------------------------

    def interval(self, period):
        if len(self) < 2:
            return None

        return (self.__ticks[1] - self.__ticks[0]) * period


    # ----------------------------------------------------------------------------------------------------------------

    def aligns(self, value):
        return bool(value in self.__ticks)


    def next(self, value):
        for tick in self.__ticks:
            if tick > value:
                return tick

        return self.__ticks[0]


    def prev(self, value):
        for tick in reversed(self.__ticks):
            if tick <= value:
                return tick

        return self.__ticks[-1]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CheckpointField:{ticks:%s}" % self.__ticks



"""
Created on 18 Sep 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A high-performance linear regression utility requiring scipy

https://www.w3schools.com/python/python_ml_linear_regression.asp
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class LinRegress(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, x, y, prec=6):
        from scipy import stats                     # late import

        if len(x) != len(y):
            raise ValueError("len x (%d) is not the same as len y (%d)" % (len(x), len(y)))

        slope, intercept, r, p, std_err = stats.linregress(x, y)

        return cls(len(x), round(slope, prec), round(intercept, prec), round(r**2, prec), round(p, prec),
                   round(std_err, prec))


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        count = jdict.get('count')
        slope = jdict.get('slope')
        intercept = jdict.get('intercept')
        r2 = jdict.get('r2')
        p = jdict.get('p')
        std_err = jdict.get('std-err')

        return cls(count, slope, intercept, r2, p, std_err)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, count, slope, intercept, r2, p, std_err):
        """
        Constructor
        """
        self.__count = count                            # int

        self.__slope = slope                            # float
        self.__intercept = intercept                    # float
        self.__r2 = r2                                  # float
        self.__p = p                                    # float
        self.__std_err = std_err                        # float


    def __len__(self):
        return self.__count


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def slope(self):
        return self.__slope


    @property
    def intercept(self):
        return self.__intercept


    @property
    def r2(self):
        return self.__r2


    @property
    def p(self):
        return self.__p


    @property
    def std_err(self):
        return self.__std_err


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['count'] = len(self)

        jdict['slope'] = self.slope
        jdict['intercept'] = self.intercept
        jdict['r2'] = self.r2
        jdict['p'] = self.p
        jdict['std-err'] = self.std_err

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "LinRegress:{slope:%s, intercept:%s, r2:%s, p:%s, std_err:%s}" %  \
               (self.slope, self.intercept, self.r2, self.p, self.std_err)



"""
Created on 1 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"calibrated-on": "2019-02-02T11:34:16Z", "offset": 50}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class SensorBaseline(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return SensorBaseline(None, 0)

        if 'calibrated_on' in jdict:                            # TODO: deprecated
            date = Datum.date(jdict.get('calibrated_on'))
            calibrated_on = LocalizedDatetime.construct_from_date(date)

        else:
            calibrated_on = LocalizedDatetime.construct_from_iso8601(jdict.get('calibrated-on'))

        offset = jdict.get('offset')

        return SensorBaseline(calibrated_on, offset)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calibrated_on, offset):
        """
        Constructor
        """
        self.__calibrated_on = calibrated_on            # LocalizedDatetime

        self.__offset = int(offset)                     # int                       ppb


    def __eq__(self, other):
        try:
            return self.calibrated_on == other.calibrated_on and self.offset == other.offset
        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['calibrated-on'] = None if self.calibrated_on is None else self.calibrated_on.as_iso8601(False)
        jdict['offset'] = self.offset

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def calibrated_on(self):
        return self.__calibrated_on


    @property
    def offset(self):
        return self.__offset


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SensorBaseline:{calibrated_on:%s, offset:%s}" %  (self.calibrated_on, self.offset)



"""
Created on 22 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an abstract machine learning model group configuration
"""

import os

from abc import ABC, abstractmethod
from collections import OrderedDict

from scs_core.data.json import PersistentJSONable

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModelConf(ABC, PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def interfaces(cls):
        return []


    @classmethod
    def is_valid_interface(cls, model_interface):
        return model_interface in cls.interfaces()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None) if skeleton else None

        uds_path = jdict.get('uds-path')
        model_interface = jdict.get('model-interface')
        model_compendium_group = jdict.get('model-compendium-group')

        return cls(uds_path, model_interface, model_compendium_group=model_compendium_group)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_path, model_interface, model_compendium_group=None):
        """
        Constructor
        """
        super().__init__()

        self.__uds_path = uds_path                                      # string
        self.__model_interface = model_interface                        # string
        self.__model_compendium_group = model_compendium_group          # string

        self._logger = Logging.getLogger()


    def __eq__(self, other):
        try:
            return self.uds_path == other.uds_path and self.model_interface == other.model_interface and \
                   self.model_compendium_group == other.model_compendium_group
        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['uds-path'] = self.uds_path
        jdict['model-interface'] = self.model_interface

        if self.model_compendium_group is not None:
            jdict['model-compendium-group'] = self.model_compendium_group

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def abs_uds_path(self, host):
        return os.path.join(host.scs_path(), self.uds_path)


    @property
    def uds_path(self):
        return self.__uds_path


    @property
    def model_interface(self):
        return self.__model_interface


    @property
    def model_compendium_group(self):
        return self.__model_compendium_group


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{uds_path:%s, model_interface:%s, model_compendium_group:%s}" %  \
               (self.uds_path, self.model_interface, self.model_compendium_group)



"""
Created on 6 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://github.com/aws/aws-iot-device-sdk-python
https://s3.amazonaws.com/aws-iot-device-sdk-python-docs/sphinx/html/index.html

https://github.com/aws/aws-iot-device-sdk-python/issues/57

https://stackoverflow.com/questions/20083858/how-to-extract-value-from-bound-method-in-python

https://github.com/aws/aws-iot-device-sdk-python-V2
"""

import logging

from AWSIoTPythonSDK.exception.AWSIoTExceptions import connectError, connectTimeoutException
from AWSIoTPythonSDK.exception.AWSIoTExceptions import disconnectError, disconnectTimeoutException

import AWSIoTPythonSDK.MQTTLib as MQTTLib

from scs_core.data.json import JSONify
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class MQTTClient(object):
    """
    classdocs
    """
    __KEEP_ALIVE_INTERVAL =         60                     # recommended: 30 default: 600 (sec)

    __PORT =                        8883

    __QUEUE_SIZE =                  -1                      # recommended: infinite
    __QUEUE_DROP_BEHAVIOUR =        MQTTLib.DROP_OLDEST     # not required for infinite queue
    __QUEUE_DRAINING_FREQUENCY =    2                       # recommended: 2 (Hz)

    __RECONN_BASE =                 1                       # recommended: 1 (sec)
    __RECONN_MAX =                  32                      # recommended: 32 or 128 (sec), was 10
    __RECONN_STABLE =               20                      # recommended: 20 (sec), was 10

    __DISCONNECT_TIMEOUT =          10                      # recommended: 10 (sec), was 40
    __OPERATION_TIMEOUT =           5                       # recommended: 5 (sec), was 20

    __PUB_QOS =                     1
    __SUB_QOS =                     1


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def __assert_logger(level):
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)

        logger = logging.getLogger("AWSIoTPythonSDK.core")
        logger.setLevel(level)
        logger.addHandler(stream_handler)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, *subscribers):
        """
        Constructor
        """
        self.__client = None
        self.__subscribers = subscribers


    # ----------------------------------------------------------------------------------------------------------------

    def connect(self, auth, debug):
        # logging...
        if debug:
            self.__assert_logger(logging.DEBUG)

        # client...
        self.__client = MQTTLib.AWSIoTMQTTClient(auth.client_id)

        # configuration...
        self.__client.configureEndpoint(auth.endpoint, self.__PORT)

        self.__client.configureCredentials(auth.root_ca_file_path, auth.private_key_path, auth.certificate_path)

        self.__client.configureAutoReconnectBackoffTime(self.__RECONN_BASE, self.__RECONN_MAX, self.__RECONN_STABLE)

        self.__client.configureOfflinePublishQueueing(self.__QUEUE_SIZE)
        self.__client.configureDrainingFrequency(self.__QUEUE_DRAINING_FREQUENCY)

        self.__client.configureConnectDisconnectTimeout(self.__DISCONNECT_TIMEOUT)
        self.__client.configureMQTTOperationTimeout(self.__OPERATION_TIMEOUT)

        # subscriptions...
        for subscriber in self.__subscribers:
            self.__client.subscribe(subscriber.topic, self.__SUB_QOS, subscriber.handler)

        # connect...
        try:
            return self.__client.connect(self.__KEEP_ALIVE_INTERVAL)

        except (connectError, connectTimeoutException) as ex:
            raise OSError(ex.__class__.__name__)


    def disconnect(self):
        if self.__client is None:
            return

        try:
            self.__client.disconnect()

        except (disconnectError, disconnectTimeoutException):
            pass

        self.__client = None


    # --------------------------------------------------------------------------------------------------------------------

    def publish(self, publication):
        if not self.__client:
            raise IOError("publish: no client")

        payload = JSONify.dumps(publication.payload)

        return self.__client.publish(publication.topic, payload, self.__PUB_QOS)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTClient:{subscribers:%s}" % Str.collection(self.__subscribers)


# --------------------------------------------------------------------------------------------------------------------

class MQTTSubscriber(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, handler):
        """
        Constructor
        """
        self.__topic = topic
        self.__handler = handler


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def handler(self):
        return self.__handler


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTSubscriber:{topic:%s, handler:%s}" % (self.topic, self.handler.__self__)



"""
Created on 03 Dec 2020

@author: Jade Page (Jade.Page@southcoastscience.com)
"""
import logging

from urllib.parse import urlencode

from scs_core.data.json import JSONify
from scs_core.data.datetime import LocalizedDatetime
from scs_core.aws.manager.lambda_message_manager import MessageManager


class AWSMessages(object):
    __REQUEST_PATH = "/topicMessages"
    __END_POINT = "aws.southcoastscience.com"
    def __init__(self, api_auth, topic, start, end, max_lines):
        """
        Constructor
        """
        self.__api_auth = api_auth
        self.__start = start
        self.__end = end
        self.__topic = topic
        self.__message_manager = None
        self.__max_lines = max_lines
        self.__message_manager = MessageManager(self.__api_auth)

    def next_url(self, time):
        next_params = {
            'topic': self.__topic,
            'startTime': LocalizedDatetime.construct_from_iso8601(time),
            'endTime': self.__end.as_iso8601()
        }
        query = urlencode(next_params)

        url = 'https://{}{}?{}'.format(self.__END_POINT, self.__REQUEST_PATH, query)

        return url

    def run(self):
        res = []
        output_count = 0
        logging.debug(("aws_messages: start: %s" % self.__start))
        logging.debug(("aws_messages: end: %s" % self.__end))
        logging.debug(("aws_messages: topic: %s" % self.__topic))
        logging.debug(("aws_messages: max_lines: %s" % self.__max_lines))
        for message in self.__message_manager.find_for_topic(self.__topic, self.__start, self.__end, False):
            if output_count < self.__max_lines:
                res.append(message)
                output_count += 1
            else:
                next_time = message.payload.get("rec")
                next_url = self.next_url(next_time)
                jstr = JSONify.dumps(res)
                return jstr, next_url

        return None, None





"""
Created on 01 Dec 2020

@author: Jade Page (Jade.Page@southcoastscience.com)
"""
import json
import logging

from urllib.parse import urlencode

from scs_core.aws.manager.topic_message_manager import MessageManager
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.path_dict import PathDict
from scs_core.data.aggregate import Aggregate
from scs_core.data.checkpoint_generator import CheckpointGenerator


class AWSAggregator(object):
    __REQUEST_PATH = "/topicMessages"
    __END_POINT = "aws.southcoastscience.com"

    def __init__(self, lambda_client, topic, start, end, checkpoint, max_lines, min_max):
        """
        Constructor
        """
        self.__lambda_client = lambda_client
        self.__start = start
        self.__end = end
        self.__topic = topic
        self.__checkpoint = checkpoint
        self.__generator = None
        self.__aggregate = None
        self.__reporter = None
        self.__max_lines = max_lines
        self.__min_max = min_max
        self.__message_manager = None



    def setup(self):
        self.__generator = CheckpointGenerator.construct(self.__checkpoint)
        self.__aggregate = Aggregate(self.__min_max, "rec", None)
        self.__message_manager = MessageManager(self.__lambda_client)

    def next_url(self, checkpoint):
        next_params = {
            'topic': self.__topic,
            'startTime': checkpoint.as_iso8601(False),
            'endTime': self.__end.as_iso8601(False),
        }
        query = urlencode(next_params)

        url = 'https://{}{}?{}'.format(self.__END_POINT, self.__REQUEST_PATH, query)

        return url

    def run(self):
        logging.debug(("aws_aggregate: start: %s" % self.__start))
        logging.debug(("aws_aggregate: end: %s" % self.__end))
        logging.debug(("aws_aggregate: topic: %s" % self.__topic))
        logging.debug(("aws_aggregate: max_lines: %s" % self.__max_lines))
        logging.debug(("aws_aggregate: min_max: %s" % self.__min_max))

        checkpoint = None
        prev_rec = None
        res = []

        document_count = 0
        output_count = 0
        processed_count = 0
        for message in self.__message_manager.find_for_topic(self.__topic, self.__start, self.__end):
            logging.debug("Got message")
            jstr = json.dumps(message.payload)
            datum = PathDict.construct_from_jstr(jstr)

            try:
                rec_node = datum.node("rec")
            except KeyError:
                continue

            document_count += 1
            logging.debug(("aws_aggregate: document_count: %s" % document_count))
            logging.debug(("aws_aggregate: document: %s" % datum))

            rec = LocalizedDatetime.construct_from_iso8601(rec_node)

            # set checkpoint...
            if checkpoint is None:
                checkpoint = self.__generator.next_localised_datetime(rec)

            # report and reset...
            if rec > checkpoint:
                result = self.__aggregate.pass_back(checkpoint)
                logging.debug(result)
                res.append(result)
                output_count += 1
                logging.debug(("aws_aggregate: output_count: %s" % output_count))

                self.__aggregate.reset()
                logging.debug("aws_aggregate: aggregate reset")

                checkpoint = self.__generator.enclosing_localised_datetime(rec)

            # duplicate recs?...
            if rec == prev_rec:
                logging.debug(("sample_aggregate: discarding duplicate: %s" % message.strip()))
                continue

            # append sample...
            self.__aggregate.append(rec, datum)

            prev_rec = rec
            processed_count += 1

            if output_count >= self.__max_lines:
                return res, self.next_url(checkpoint)

        return res, self.next_url(checkpoint)




"""
Created on 2 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class AbstractTopic(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, name, description, is_public, info):
        """
        Constructor
        """
        self.__path = path                          # string
        self.__name = name                          # string
        self.__description = description            # string

        self.__is_public = is_public                # bool

        self.__info = info                          # TopicInfo


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.path is not None:
            jdict['topic'] = self.path

        jdict['name'] = self.name
        jdict['description'] = self.description

        jdict['public'] = self.is_public

        jdict['topic-info'] = self.info

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def path(self):
        return self.__path


    @property
    def name(self):
        return self.__name


    @property
    def description(self):
        return self.__description


    @property
    def is_public(self):
        return self.__is_public


    @property
    def info(self):
        return self.__info



"""
Created on 14 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://aws.amazon.com/premiumsupport/knowledge-center/lambda-function-idempotent/

document example:
{"tag": "scs-bgx-003", "rec": "2021-04-14T08:49:22+01:00", "message-rec": "2021-04-14T08:49:20+01:00",
"result": "ERROR", "context": "stderr output"}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationCheck(JSONable):
    """
    classdocs
    """

    RESULT_NO_RESPONSE =                "NO RESPONSE"
    RESULT_ERROR =                      "ERROR"
    RESULT_MALFORMED =                  "MALFORMED:"
    RESULT_MALFORMED_SAMPLE =           "MALFORMED:SAMPLE"
    RESULT_MALFORMED_CONFIG =           "MALFORMED:CONFIG"
    RESULT_NOT_SUPPORTED =              "NOT SUPPORTED"
    RESULT_RECEIVED =                   "RECEIVED:"
    RESULT_RECEIVED_NEW =               "RECEIVED:NEW"
    RESULT_RECEIVED_UNCHANGED =         "RECEIVED:UNCHANGED"
    RESULT_RECEIVED_UPDATED =           "RECEIVED:UPDATED"

    __RESULTS = {
        'NOR': RESULT_NO_RESPONSE,
        'ERR': RESULT_ERROR,
        'M': RESULT_MALFORMED,
        'MSA': RESULT_MALFORMED_SAMPLE,
        'MCO': RESULT_MALFORMED_CONFIG,
        'NSP': RESULT_NOT_SUPPORTED,
        'R': RESULT_RECEIVED,
        'RNW': RESULT_RECEIVED_NEW,
        'RUN': RESULT_RECEIVED_UNCHANGED,
        'RUP': RESULT_RECEIVED_UPDATED
    }

    @classmethod
    def result_codes(cls):
        return cls.__RESULTS.keys()


    @classmethod
    def result_string(cls, code):
        if code is None:
            return None

        return cls.__RESULTS[code]


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        message_rec = LocalizedDatetime.construct_from_iso8601(jdict.get('message-rec'))
        result = jdict.get('result')
        context = jdict.get('context')

        return cls(tag, rec, message_rec, result, context=context)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, message_rec, result, context=''):
        """
        Constructor
        """
        self.__tag = tag                        # string
        self.__rec = rec                        # LocalizedDatetime - when this record was created
        self.__message_rec = message_rec        # LocalizedDatetime - the rec of the ControlReceipt
        self.__result = result                  # string
        self.__context = context                # string


    def __lt__(self, other):
        if self.tag < other.tag:
            return True

        if self.tag > other.tag:
            return False

        if self.rec < other.rec:
            return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['rec'] = None if self.rec is None else self.rec.as_iso8601()
        jdict['message-rec'] = None if self.message_rec is None else self.message_rec.as_iso8601()
        jdict['result'] = self.result
        jdict['context'] = self.context if self.context else ["-"]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def rec(self):
        return self.__rec


    @property
    def message_rec(self):
        return self.__message_rec


    @property
    def result(self):
        return self.__result


    @property
    def context(self):
        return self.__context


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationCheck:{tag:%s, rec:%s, message_rec:%s, result:%s, context:%s}" % \
               (self.tag, self.rec, self.message_rec, self.result, self.context)



"""
Created on 4 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

AQCSV: Qualifiers

NB: initialisation is performed at the foot of this class

example:
{"code": "BL", "description": "QA audit", "type-code": "ND", "type-description": "Null Data"}

https://www.airnow.gov/
"""

import os

from collections import OrderedDict

from scs_core.csv.csv_archive import CSVArchive
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Qualifier(CSVArchive, JSONable):
    """
    classdocs
    """

    _retrieved = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def archive_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'archive', 'qualifiers.csv')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        code = jdict.get('code')
        description = jdict.get('description')

        type_code = jdict.get('type-code')
        type_description = str(jdict.get('type-description'))

        return Qualifier(code, description, type_code, type_description)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, code, description, type_code, type_description):
        """
        Constructor
        """
        self.__code = code                                      # string
        self.__description = description                        # string

        self.__type_code = type_code                            # string
        self.__type_description = type_description              # string


    def __eq__(self, other):
        try:
            return self.code == other.code and self.description == other.description and \
                   self.type_code == other.type_code and self.type_description == other.type_description

        except AttributeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['code'] = self.code
        jdict['description'] = self.description

        jdict['type-code'] = self.type_code
        jdict['type-description'] = self.type_description

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.code


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def code(self):
        return self.__code


    @property
    def description(self):
        return self.__description


    @property
    def type_code(self):
        return self.__type_code


    @property
    def type_description(self):
        return self.__type_description


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Qualifier:{code:%s, description:%s, type_code:%s, type_description:%s}" % \
               (self.code, self.description, self.type_code, self.type_description)


# --------------------------------------------------------------------------------------------------------------------
# initialisation...

Qualifier.retrieve()



"""
Created on 25 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://howchoo.com/g/ywi5m2vkodk/working-with-datetime-objects-and-timezones-in-python

example JSON:
{"path": "val.CO.cnc", "index": 0, "value": 340,
"sample": {"rec": "2021-07-30T19:45:00Z", "val": {
"NO2": {"weV": 0.31812, "cnc": 57.0, "aeV": 0.28576, "weC": 0.00323},
"Ox": {"weV": 0.41573, "cnc": 24.4, "aeV": 0.40003, "weC": 0.00181},
"NO": {"weV": 0.3081, "cnc": 34.9, "aeV": 0.28073, "weC": 0.05806},
"CO": {"weV": 0.35202, "cnc": 339.8, "aeV": 0.2881, "weC": 0.06908},
"sht": {"hmd": 69.9, "tmp": 18.5}}, "tag": "scs-bgx-401",
"exg": {"vB20": {"NO2": {"cnc": 14.4}}}}}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict

from scs_core.gas.a4.a4 import A4
from scs_core.gas.afe_baseline import AFEBaseline
from scs_core.gas.sensor_baseline import SensorBaseline


# --------------------------------------------------------------------------------------------------------------------

class Minimum(JSONable):
    """
    classdocs
    """

    FIELD_SELECTIONS = {'V': 'val.', 'E': 'exg.'}           # Val (model input) or Exg (model output)

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def init(cls):
        A4.init()


    @classmethod
    def find_minimums(cls, data, field_selection):
        if not data:
            return []

        # minimums...
        field_group = cls.FIELD_SELECTIONS[field_selection]

        minimums = {path: None for path in PathDict(data[0]).paths()
                    if path.startswith(field_group) and (path.endswith('.cnc') or path.endswith('.vCal'))}

        # data...
        for i in range(len(data)):
            datum = PathDict(data[i])

            for path in minimums:
                value = round(float(datum.node(path)))

                if minimums[path] is None or value < minimums[path].value:
                    minimums[path] = cls(path, i, value, data[i])

        return [minimums[path] for path in sorted(minimums.keys())]


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, index, value, sample):
        """
        Constructor
        """
        self.__path = path                          # string
        self.__index = int(index)                   # int
        self.__value = int(round(value))            # int
        self.__sample = sample                      # dict (of GasesSample)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['path'] = self.path
        jdict['index'] = self.index
        jdict['value'] = self.value
        jdict['sample'] = self.sample

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def update_already_done(self, configuration, end):
        cmd = self.__cmd()

        if cmd == 'scd30_baseline':
            baseline = configuration.scd30_baseline
            reported_baseline = None if baseline is None else baseline.sensor_baseline

        elif cmd == 'afe_baseline':
            baseline = AFEBaseline.null_datum() if configuration.afe_baseline is None else configuration.afe_baseline
            reported_baseline = baseline.sensor_baseline(configuration.afe_id.sensor_index(self.gas))

        elif cmd == 'vcal_baseline':
            baseline = configuration.vcal_baseline
            reported_baseline = None if baseline is None else baseline.sensor_baseline(self.gas)

        elif cmd == 'gas_baseline':
            baseline = configuration.gas_baseline
            reported_baseline = None if baseline is None else baseline.sensor_baseline(self.gas)

        else:
            raise ValueError(cmd)

        sensor_baseline = SensorBaseline(None, 0) if reported_baseline is None else reported_baseline

        return sensor_baseline.calibrated_on is not None and sensor_baseline.calibrated_on > end


    def cmd_tokens(self, conf_minimums):
        cmd = self.__cmd()

        if cmd == 'scd30_baseline':
            return [cmd, '-vc', conf_minimums[self.gas], self.value]

        if cmd == 'afe_baseline':
            return [cmd, '-vc', self.gas, conf_minimums[self.gas], self.value]

        if cmd == 'vcal_baseline':
            return [cmd, '-vs', self.gas, conf_minimums[self.gas] - self.value]     # vCal does not hold its correction

        if cmd == 'gas_baseline':
            return [cmd, '-vc', self.gas, conf_minimums[self.gas], self.value]

        raise ValueError(cmd)


    # ----------------------------------------------------------------------------------------------------------------

    def __cmd(self):
        if self.gas == 'CO2':
            return 'scd30_baseline'

        if self.source == 'val' and self.interpretation == 'cnc':
            return 'afe_baseline'

        if self.source == 'val' and self.interpretation == 'vCal':
            return 'vcal_baseline'

        if self.source == 'exg':
            return 'gas_baseline'

        raise ValueError(self.source)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def source(self):
        return self.__path.split('.')[0]


    @property
    def interpretation(self):
        return self.__path.split('.')[-1]


    @property
    def gas(self):
        return self.__path.split('.')[-2]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def path(self):
        return self.__path


    @property
    def index(self):
        return self.__index


    @property
    def value(self):
        return self.__value


    @value.setter
    def value(self, value):
        self.__value = int(round(value))


    @property
    def sample(self):
        return self.__sample


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Minimum:{path:%s, index:%s, value:%s, sample:%s}" % \
               (self.path, self.index, self.value, self.sample)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
  {
    "client-id": "5891",
    "description": "BB development platform",
    "name": "Praxis/BGB 000000",
    "tags": [
      "pm1",
      "pm2.5",
      "pm10",
      "co",
      "no",
      "no2",
      "o3",
      "temperature",
      "humidity"
    ],
    "location": {
      "lat": 50.82313,
      "lon": -0.122922,
      "postcode": "bn2 0da"
    }
  },
"""

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.osio.data.location import Location


# --------------------------------------------------------------------------------------------------------------------

class DeviceSummary(JSONable):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        client_id = jdict.get('client-id')
        name = jdict.get('name')
        description = jdict.get('description')

        location = Location.construct_from_jdict(jdict.get('location'))

        tags = jdict.get('tags')

        device = DeviceSummary(client_id, name, description, location, tags)

        return device


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client_id, name, description, location, tags):
        """
        Constructor
        """
        self.__client_id = client_id                        # string (int by convention)
        self.__name = name                                  # string
        self.__description = description                    # string

        self.__location = location                          # Location

        self.__tags = tags                                  # array of string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.client_id is not None:
            jdict['client-id'] = self.client_id

        jdict['name'] = self.name
        jdict['description'] = self.description

        if self.location is not None:
            jdict['location'] = self.location

        jdict['tags'] = self.tags

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def client_id(self):
        return self.__client_id


    @property
    def name(self):
        return self.__name


    @property
    def description(self):
        return self.__description


    @property
    def location(self):
        return self.__location


    @property
    def tags(self):
        return self.__tags


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceSummary:{client_id:%s, name:%s, description:%s, location:%s, tags:%s}" % \
                    (self.client_id, self.name, self.description, self.location, self.tags)



"""
Created on 17 Jun 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict
from enum import Enum
from http import HTTPStatus

from scs_core.aws.data.alert import AlertStatus
from scs_core.aws.data.http_response import HTTPResponse

from scs_core.data.datum import Datum
from scs_core.data.str import Str

from scs_core.sys.http_exception import HTTPException


# --------------------------------------------------------------------------------------------------------------------

class AlertStatusFinder(object):
    """
    classdocs
    """

    __URL = "https://nssnahspkj.execute-api.us-west-2.amazonaws.com/default/AlertHistoryLatestFinder"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client, auth):
        self.__http_client = http_client
        self.__auth = auth


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, id_filter, cause_filter, response_mode):
        request = AlertStatusFinderRequest(id_filter, cause_filter, response_mode)
        headers = {'Authorization': self.__auth.email_address}

        response = self.__http_client.get(self.__URL, headers=headers, params=request.params())

        return AlertStatusFinderResponse.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertStatusFinder:{auth:%s}" % self.__auth


# --------------------------------------------------------------------------------------------------------------------

class AlertStatusFinderRequest(object):
    """
    classdocs
    """

    MODE = Enum('Mode', 'HISTORY LATEST')

    ID_FILTER = 'id'
    CAUSE_FILTER = 'cause'
    RESPONSE_MODE = 'responseMode'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        id_filter = qsp.get(cls.ID_FILTER)
        cause_filter = qsp.get(cls.CAUSE_FILTER)

        try:
            response_mode = cls.MODE[qsp.get(cls.RESPONSE_MODE)]
        except KeyError:
            response_mode = None

        return cls(id_filter, cause_filter, response_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id_filter, cause_filter, response_mode):
        """
        Constructor
        """
        self.__id_filter = Datum.int(id_filter)                     # int
        self.__cause_filter = cause_filter                          # string
        self.__response_mode = response_mode                        # enum


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.id_filter is None and self.cause_filter is None:
            return False

        if self.response_mode is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {
            self.ID_FILTER: self.id_filter,
            self.CAUSE_FILTER: self.cause_filter,
            self.RESPONSE_MODE: self.response_mode.name
        }

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id_filter(self):
        return self.__id_filter


    @property
    def cause_filter(self):
        return self.__cause_filter


    @property
    def response_mode(self):
        return self.__response_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertStatusFinderRequest:{id_filter:%s, cause_filter:%s, response_mode:%s}" % \
               (self.id_filter, self.cause_filter, self.response_mode)


# --------------------------------------------------------------------------------------------------------------------

class AlertStatusFinderResponse(HTTPResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # print("jdict: %s" % jdict)

        status = HTTPStatus(jdict.get('statusCode'))

        if status != HTTPStatus.OK:
            raise HTTPException(status.value, status.phrase, status.description)

        mode = AlertStatusFinderRequest.MODE[jdict.get('mode')]

        alert_statuses = []
        if jdict.get('alert-statuses'):
            for alert_jdict in jdict.get('alert-statuses'):
                alert_statuses.append(AlertStatus.construct_from_jdict(alert_jdict))

        next_url = jdict.get('next')

        return cls(status, mode, alert_statuses, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, status, mode, alert_statuses, next_url=None):
        """
        Constructor
        """
        super().__init__(status)

        self.__mode = mode                              # AlertStatusRequest.Mode member
        self.__alert_statuses = alert_statuses          # array of AlertStatus
        self.__next_url = next_url                      # URL string


    def __len__(self):
        return len(self.alert_statuses)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['statusCode'] = self.status.value

        if self.mode is not None:
            jdict['mode'] = self.mode.name

        if self.alert_statuses is not None:
            jdict['alert-statuses'] = self.alert_statuses

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def mode(self):
        return self.__mode


    @property
    def alert_statuses(self):
        return self.__alert_statuses


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AlertStatusFinderResponse:{status:%s, mode:%s, alert_statuses:%s, next_url:%s}" % \
               (self.status, self.mode, Str.collection(self.alert_statuses), self.next_url)



"""
Created on 17 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

http://pythoncentral.io/hashing-strings-with-python/

example document:
{"tag": "scs-be2-3", "attn": "bruno",
"rec": "2017-08-29T10:12:32.056+00:00",
"cmd": {"cmd": "?", "params": [], "stdout": ["[\"afe_baseline\", \"afe_calib\""], "stderr": [], "ret": 0},
"omd": "6e81c77aa20562ea06e0e32158d2c7c9431ed251cd5790917f6cb385f6cf62c0",
"digest": "59cb897c308050f6d07f400cc3b784bfa81938d4e8a36896bec2126d24c8fe00"}
"""

import hashlib

from collections import OrderedDict

from scs_core.control.command import Command

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable, JSONify

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class ControlReceipt(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        tag = jdict.get('tag')
        attn = jdict.get('attn')

        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        command = Command.construct_from_jdict(jdict.get('cmd'))
        omd = jdict.get('omd')
        digest = jdict.get('digest')

        datum = cls(tag, attn, rec, command, omd, digest)

        return datum


    @classmethod
    def construct_from_datum(cls, datum, rec, command, key):
        digest = ControlReceipt.__hash(datum.attn, rec, command, datum.digest, key)

        return cls(datum.attn, datum.tag, rec, command, datum.digest, digest)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __hash(cls, tag, rec, command, omd, key):
        rec_iso8601 = rec.as_iso8601(include_millis=Sample.INCLUDE_MILLIS)

        text = str(tag) + JSONify.dumps(rec_iso8601) + JSONify.dumps(command) + str(omd) + str(key)
        hash_object = hashlib.sha256(text.encode())

        return hash_object.hexdigest()


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, attn, rec, command, omd, digest):
        """
        Constructor
        """
        self.__tag = tag                    # string
        self.__attn = attn                  # string

        self.__rec = rec                    # LocalizedDatetime
        self.__command = command            # Command
        self.__omd = omd                    # string
        self.__digest = digest              # string


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self, key):
        digest = self.__hash(self.tag, self.rec, self.command, self.omd, key)

        return digest == self.__digest


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['attn'] = self.attn

        jdict['rec'] = self.rec.as_iso8601(include_millis=Sample.INCLUDE_MILLIS)

        jdict['cmd'] = self.command
        jdict['omd'] = self.omd
        jdict['digest'] = self.__digest

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def attn(self):
        return self.__attn


    @property
    def rec(self):
        return self.__rec


    @property
    def command(self):
        return self.__command


    @property
    def omd(self):
        return self.__omd


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ControlReceipt:{tag:%s, attn:%s, rec:%s, command:%s, omd:%s, digest:%s}" % \
               (self.tag, self.attn, self.rec, self.command, self.omd, self.__digest)



"""
Created on 30 Sep 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://medium.com/@houzier.saurav/aws-cognito-with-python-6a2867dd02c6

A single manager for all required cognito related functions.

"""
# --------------------------------------------------------------------------------------------------------------------
import ast
import base64
import hashlib
import hmac
import json

# --------------------------------------------------------------------------------------------------------------------
import logging


class CognitoManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, cognito_client, app_id, app_id_secret, pool_id):
        self.__cognito_client = cognito_client
        self.__app_id = app_id
        self.__app_id_secret = app_id_secret
        self.__pool_id = pool_id

    # ----------------------------------------------------------------------------------------------------------------
    # Main functionality
    # ----------------------------------------------------------------------------------------------------------------

    def initiate_auth(self, user, password):
        hash_key = self.generate_hash(user)
        res = self.__cognito_client.admin_initiate_auth(
            UserPoolId=self.__pool_id,
            ClientId=self.__app_id,
            AuthFlow='ADMIN_NO_SRP_AUTH',
            AuthParameters={
                'USERNAME': user,
                'SECRET_HASH': hash_key,
                'PASSWORD': password,
            },
            ClientMetadata={
                'username': user,
                'password': password,
            })

        return res

    def generate_hash(self, user):
        message = user + self.__app_id
        digest1 = hmac.new(str(self.__app_id_secret).encode('utf-8'),
                           msg=str(message).encode('utf-8'), digestmod=hashlib.sha256).digest()
        digest2 = base64.b64encode(digest1).decode()
        return digest2

    # ----------------------------------------------------------------------------------------------------------------
    # Admin functionality
    # ----------------------------------------------------------------------------------------------------------------

    def create_account(self, user, email, name, surname, admin=False):
        res = self.__cognito_client.admin_create_user(
            UserPoolId=self.__pool_id,
            Username=user,
            UserAttributes=[
                {"Name": "given_name", "Value": name},
                {"Name": "family_name", "Value": surname},
                {"Name": "email", "Value": email},
                {"Name": "custom:super", "Value": str(admin)},
            ],
            DesiredDeliveryMediums=[
                'EMAIL',
            ],
            MessageAction='SUPPRESS'
        )

        return res

    def set_password(self, user, password):
        res = self.__cognito_client.admin_set_user_password(
            UserPoolId=self.__pool_id,
            Username=user,
            Password=password,
            Permanent=True
        )

        return res

    def delete_account(self, user):
        res = self.__cognito_client.admin_delete_user(
            UserPoolId=self.__pool_id,
            Username=user
        )

        return res

        # CognitoIdentityProvider.Client.exceptions.UserNotFoundException

    def password_reset(self, user):
        res = self.__cognito_client.admin_reset_user_password(
            UserPoolId=self.__pool_id,
            Username=user
        )

        return res

    # CognitoIdentityProvider.Client.exceptions.TooManyRequestsException
    # CognitoIdentityProvider.Client.exceptions.NotAuthorizedException
    # CognitoIdentityProvider.Client.exceptions.UserNotFoundException

    def get_user_information(self, user):
        res = self.__cognito_client.admin_get_user(
            UserPoolId=self.__pool_id,
            Username=user
        )

        return res

    def make_super(self, user):
        res = self.__cognito_client.admin_update_user_attributes(
            UserPoolId=self.__pool_id,
            Username=user,
            UserAttributes=[
                {
                    'Name': 'custom:super',
                    'Value': 'True'
                }]
        )
        return res

    def is_super(self, user):
        b_is_admin = False
        x = self.get_user_information(user)
        for item in x["UserAttributes"]:
            if item["Name"] == "custom:super":
                b_is_admin = item["Value"]
                break

        return ast.literal_eval(b_is_admin)

    # ----------------------------------------------------------------------------------------------------------------
    # User functionality
    # ----------------------------------------------------------------------------------------------------------------

    def user_sign_up(self, user, password, name, surname, email, admin=False):
        res = self.__cognito_client.sign_up(
            ClientId=self.__app_id,
            SecretHash=self.generate_hash(user),
            Username=user,
            Password=password,
            UserAttributes=[
                {"Name": "given_name", "Value": name},
                {"Name": "family_name", "Value": surname},
                {"Name": "email", "Value": email},
                {"Name": "custom:super", "Value": str(admin)},
            ]
        )

        return res


    def password_change(self, token, old_pass, new_pass):
        res = self.__cognito_client.change_password(
            PreviousPassword=old_pass,
            ProposedPassword=new_pass,
            AccessToken=token
        )

        return res

    def admin_password_change(self, user, new_pass):
        res = self.__cognito_client.admin_set_user_password(
            UserPoolId=self.__pool_id,
            Username=user,
            Password=new_pass,
            Permanent=True
        )

        return res


# --------------------------------------------------------------------------------------------------------------------

class CognitoCredentials(object):
    """
    classdocs
    """
    PASSWORD = 'password'
    ADMIN = 'is_admin'
    BODY_USER = 'username'
    EVENT_USER = 'cognito:username'
    EVENT_EMAIL = 'email'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_body(cls, body):
        if not body:
            return None

        jdict = json.loads(body)

        username = jdict[cls.BODY_USER]
        password = jdict[cls.PASSWORD]

        return cls(username, password)

    @classmethod
    def construct_from_event(cls, event):
        if not event:
            return None

        if not event["requestContext"]["authorizer"]["claims"]:
            return None

        data = event["requestContext"]["authorizer"]["claims"]
        logging.warning(data)
        username = data[cls.EVENT_USER]
        password = None
        email = data[cls.EVENT_EMAIL]

        return cls(username, password, False, email)

    def __init__(self, username, password, is_admin=False, email=None):
        """
        Constructor
        """
        self.__username = username  # string
        self.__password = password  # string
        self.__email = email  # string
        self.__is_admin = bool(is_admin)  # string > bool

    @property
    def username(self):
        return self.__username

    @property
    def password(self):
        return self.__password

    @property
    def is_admin(self):
        return self.__is_admin

    @property
    def email(self):
        return self.__email

    @is_admin.setter
    def is_admin(self, value):
        self.__is_admin = value

    def __str__(self, *args, **kwargs):
        return "CognitoCredentials:{username:%s, password:%s, email:%s, admin:%s}" % self.username, \
               self.password, self.email, self.is_admin


# --------------------------------------------------------------------------------------------------------------------

class CognitoAccount(object):
    """
    classdocs
    """
    USER = "username"
    EMAIL = "email"
    FORENAME = "forename"
    SURNAME = "surname"
    PASSWORD = "password"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_body(cls, body):
        if not body:
            return None

        jdict = json.loads(body)

        username = jdict[cls.USER]
        password = jdict[cls.PASSWORD]
        email = jdict[cls.EMAIL]
        forename = jdict[cls.FORENAME]
        surname = jdict[cls.SURNAME]


        return cls(username, email, forename, surname, password)

    def __init__(self, username, email, forename, surname, password, admin=False):
        """
        Constructor
        """
        self.__username = username  # string
        self.__forename = forename  # string
        self.__email = email  # string
        self.__admin = bool(admin)  # string > bool
        self.__surname = surname
        self.__password = password

    @property
    def username(self):
        return self.__username

    @property
    def forename(self):
        return self.__forename

    @property
    def surname(self):
        return self.__surname

    @property
    def admin(self):
        return self.__admin

    @property
    def password(self):
        return self.__password

    @property
    def email(self):
        return self.__email

    def __str__(self, *args, **kwargs):
        return "CognitoAccount:{username:%s, forename:%s, surname:%s, email:%s, admin:%s}" % self.username, \
               self.forename, self.surname, self.email, self.admin




"""
Created on 11 Jan 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json
import time

from AWSIoTPythonSDK.exception.operationError import operationError
from AWSIoTPythonSDK.exception.operationTimeoutException import operationTimeoutException

from scs_core.control.control_datum import ControlDatum
from scs_core.control.control_receipt import ControlReceipt

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.publication import Publication

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ControlHandler(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host_tag, device_tag):
        """
        Constructor
        """
        self.__host_tag = host_tag                      # string
        self.__device_tag = device_tag                  # string

        self.__outgoing_pub = None                      # ControlDatum
        self.__receipt = None                           # ControlReceipt

        self.__logger = Logging.getLogger()             # Logger


    # ----------------------------------------------------------------------------------------------------------------

    def publish(self, client, topic, cmd_tokens, cmd_timeout, key):
        now = LocalizedDatetime.now().utc()
        str_tokens = [str(token) for token in cmd_tokens]

        # datum...
        datum = ControlDatum.construct(self.__host_tag, self.__device_tag, now, str_tokens, cmd_timeout, key)

        publication = Publication(topic, datum)
        self.set_outgoing(publication)

        self.__logger.info(datum)

        # publish...
        try:
            success = client.publish(publication)
            self.__logger.info("paho: %s" % "1" if success else "0")

        except (OSError, operationError, operationTimeoutException) as ex:
            self.__logger.error(ex.__class__.__name__)
            exit(1)

        # subscribe...
        timeout = time.time() + cmd_timeout

        while True:
            if self.receipt:
                if not self.receipt.is_valid(key):
                    raise ValueError("invalid digest: %s" % self.receipt)

                self.__logger.info(self.receipt)

                return self.receipt.command.stdout, self.receipt.command.stderr

            if time.time() > timeout:
                raise TimeoutError(cmd_timeout)

            time.sleep(0.1)


    def set_outgoing(self, outgoing_pub):
        self.__outgoing_pub = outgoing_pub
        self.__receipt = None


    def handle(self, _client, _userdata, message):
        payload = json.loads(message.payload.decode())

        try:
            receipt = ControlReceipt.construct_from_jdict(payload)
        except TypeError:
            return

        if receipt.tag != self.__device_tag:
            return

        if receipt.attn is not None and receipt.attn != self.__host_tag:
            return

        if receipt.omd != self.__outgoing_pub.payload.digest:
            return

        self.__receipt = receipt


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def receipt(self):
        return self.__receipt


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ControlHandler:{host_tag:%s, device_tag:%s, outgoing_pub:%s, receipt:%s}" %  \
               (self.__host_tag, self.__device_tag, self.__outgoing_pub, self.receipt)



"""
Created on 8 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
  "id": "south-coast-science-dev",
  "name": "South Coast Science  (dev)",
  "website": "https://www.southcoastscience.com/",
  "description": "Development operations for South Coast Science air quality monitoring instruments.",
  "email": "bruno.beloff@southcoastscience.com"
}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Organisation(JSONable):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')
        name = jdict.get('name')
        website = jdict.get('website')
        description = jdict.get('description')
        email = jdict.get('email')

        return Organisation(id, name, website, description, email)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, name, website, description, email):
        """
        Constructor
        """
        self.__id = id                          # string
        self.__name = name                      # string
        self.__website = website                # string
        self.__description = description        # string
        self.__email = email                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.id is not None:
            jdict['id'] = self.id

        jdict['name'] = self.name
        jdict['website'] = self.website
        jdict['description'] = self.description
        jdict['email'] = self.email

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def name(self):
        return self.__name


    @property
    def website(self):
        return self.__website


    @property
    def description(self):
        return self.__description


    @property
    def email(self):
        return self.__email


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Organisation:{id:%s, name:%s, website:%s, description:%s, email:%s}" % \
               (self.id, self.name, self.website, self.description, self.email)



"""
Created on 20 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"rec": "2021-10-06T11:13:07Z", "tag": "scs-be2-3", "ver": 1.00, "val": {"tz": {"name": "Europe/London",
"utc-offset": "+01:00"}, "gps": {"pos": [null, null], "elv": null, "qual": 0},
"sch": {"scs-climate": {"interval": 60.0, "tally": 1}, "scs-gases": {"interval": 10.0, "tally": 1},
"scs-status": {"interval": 60.0, "tally": 1}}, "tmp": {"brd": 29.4}, "up": {"period": "00-00:22:00", "users": 3,
"load": {"av1": 0.02, "av5": 0.34, "av15": 0.67}}, "sig": {"quality": null, "recent": null}}}
"""

from collections import OrderedDict

from scs_core.aqcsv.conf.airnow_site_conf import AirNowSiteConf

from scs_core.data.datetime import LocalizedDatetime

from scs_core.location.timezone import Timezone

from scs_core.position.gps_datum import GPSDatum

from scs_core.sample.sample import Sample

from scs_core.sync.schedule import Schedule

from scs_core.sys.modem import Signal
from scs_core.sys.system_temp import SystemTemp
from scs_core.sys.uptime_datum import UptimeDatum


# --------------------------------------------------------------------------------------------------------------------

class StatusSample(Sample):
    """
    classdocs
    """

    VERSION = 1.0

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # Sample...
        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        version = jdict.get('ver', cls.DEFAULT_VERSION)

        val = jdict.get('val')

        # StatusSample...
        airnow = AirNowSiteConf.construct_from_jdict(val.get('airnow'))
        timezone = Timezone.construct_from_jdict(val.get('tz'))
        position = GPSDatum.construct_from_jdict(val.get('gps'))
        temperature = SystemTemp.construct_from_jdict(val.get('tmp'))
        schedule = Schedule.construct_from_jdict(val.get('sch'))
        uptime = UptimeDatum.construct_from_jdict(val.get('up'))
        modem_signal = Signal.construct_from_jdict(val.get('sig'))

        # PSUReport classes are not available to the scs_core package

        return cls(tag, rec, airnow, timezone, position, temperature, schedule, uptime, val.get('psu'), modem_signal,
                   version=version)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, airnow, timezone, position, temperature, schedule, uptime, psu_report, modem_signal,
                 version=None):
        """
        Constructor
        """
        if version is None:
            version = self.VERSION

        super().__init__(tag, rec, version)

        self.__airnow = airnow                                      # AirNowSiteConf
        self.__timezone = timezone                                  # Timezone
        self.__position = position                                  # GPSDatum
        self.__temperature = temperature                            # HostStatus
        self.__schedule = schedule                                  # Schedule
        self.__uptime = uptime                                      # UptimeDatum
        self.__psu_report = psu_report                              # PSUReport
        self.__modem_signal = modem_signal                          # Signal


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def has_invalid_value(cls):
        # TODO: implement has_invalid_value
        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        jdict = OrderedDict()

        if self.airnow is not None:
            jdict['airnow'] = self.airnow.as_json()

        if self.timezone is not None:
            jdict['tz'] = self.timezone.as_json()

        if self.position is not None:
            jdict['gps'] = self.position.as_json()

        jdict['sch'] = self.schedule.as_json()
        jdict['tmp'] = self.temperature.as_json()
        jdict['up'] = self.uptime.as_json()

        if self.psu_report is not None:
            jdict['psu'] = self.psu_report.as_json()

        if self.modem_signal is not None:
            jdict['sig'] = self.modem_signal.as_json()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def airnow(self):
        return self.__airnow


    @property
    def timezone(self):
        return self.__timezone


    @property
    def position(self):
        return self.__position


    @property
    def temperature(self):
        return self.__temperature


    @property
    def schedule(self):
        return self.__schedule


    @property
    def uptime(self):
        return self.__uptime


    @property
    def psu_report(self):
        return self.__psu_report


    @property
    def modem_signal(self):
        return self.__modem_signal


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "StatusSample:{tag:%s, rec:%s, airnow:%s, timezone:%s, position:%s, " \
               "temperature:%s, schedule:%s, uptime:%s, psu_report:%s, modem_signal:%s}" % \
            (self.tag, self.rec, self.airnow, self.timezone, self.position,
             self.temperature, self.schedule, self.uptime, self.psu_report, self.modem_signal)



"""
Created on 26 Nov 2020

@author: Jade Page (Jade.Page@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class DeviceReport(JSONable):

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        device_tag = jdict.get('device_tag')
        bylines = jdict.get('bylines')
        power = jdict.get('power')
        status = jdict.get('status')
        uptime = jdict.get('uptime')
        emails = jdict.get('emails')

        return cls(device_tag, bylines, power, status, uptime, emails)

    @classmethod
    def construct_from_monitor(cls, device_tag, byline_list, power_list, status_list, uptime_list, email_list):
        failures = 0
        bylines = None
        power = None
        status = None
        uptime = None
        emails = None

        for key, value in byline_list.items():
            if key == device_tag:
                bylines = value

        for key, value in power_list.items():
            if key == device_tag:
                power = value

        for key, value in status_list.items():
            if key == device_tag:
                status = value

        for key, value in uptime_list.items():
            if key == device_tag:
                uptime = value

        for key, value in email_list.items():
            if key == device_tag:
                emails = value

        if bylines is None:
            failures += 1
        if power is None:
            failures += 1
        if status is None:
            failures += 1
        if uptime is None:
            failures += 1

        if failures > 3:
            return None

        return cls(device_tag, bylines, power, status, uptime, emails)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_tag, bylines, power, status, uptime, emails):
        """
        Constructor
        """
        self.__device_tag = device_tag
        self.__bylines = bylines

        self.__power = power
        self.__status = status

        self.__uptime = uptime
        self.__emails = emails

    def as_json(self):
        jdict = OrderedDict()

        jdict['device_tag'] = self.device_tag
        jdict['bylines'] = self.bylines

        jdict['power'] = self.power
        jdict['status'] = self.status

        jdict['uptime'] = self.uptime
        jdict['emails'] = self.emails

        return jdict

    @property
    def device_tag(self):
        return self.__device_tag

    @property
    def bylines(self):
        return self.__bylines

    @property
    def power(self):
        return self.__power

    @property
    def emails(self):
        return self.__emails

    @property
    def status(self):
        return self.__status

    @property
    def uptime(self):
        return self.__uptime


    def __str__(self, *args, **kwargs):
        return "DeviceReport:{device_tag:%s, bylines:%s, power:%s, status:%s, uptime:%s, emails:%s}" %  \
               (self.device_tag, self.bylines, self.power, self.status, self.uptime, self.emails)




"""
Created on 30 Apr 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from subprocess import check_output, Popen, PIPE

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class Pipe(object):
    """
    classdocs
    """

    @staticmethod
    def __args(command):
        return [str(arg) for arg in command]


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, *command):
        """
        Constructor
        """
        self.__commands = command               # array of arrays of scalar


    # ----------------------------------------------------------------------------------------------------------------

    def wait(self):
        if not self.__commands:
            return None

        count = len(self.__commands)

        p = None
        p_in = None

        for i in range(count):
            p_out = PIPE if i < count - 1 else None
            p = Popen(self.__args(self.__commands[i]), stdin=p_in, stdout=p_out)
            p_in = p.stdout

        p.wait()

        return p.returncode


    def check_output(self):
        if not self.__commands:
            return None

        count = len(self.__commands) - 1

        p_in = None

        for i in range(count):
            p = Popen(self.__args(self.__commands[i]), stdin=p_in, stdout=PIPE)
            p_in = p.stdout

        output = check_output(self.__args(self.__commands[count]), stdin=p_in)

        return output.decode().strip()


    # ----------------------------------------------------------------------------------------------------------------

    def as_script(self):
        script = ''

        for command in self.__commands:
            if script:
                script += ' | '

            for arg in self.__args(command):
                script += arg + ' '

        return script.strip()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Pipe:{commands:%s}" % Str.collection(self.__commands)



"""
Created on 16 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

requires pycryptodome

https://stackoverflow.com/questions/42568262/how-to-encrypt-text-with-a-password-in-python
https://github.com/openthread/openthread/issues/1137
"""

from base64 import b64decode, b64encode

from Crypto import Random
from Crypto.Cipher import AES
from Crypto.Hash import SHA256


# --------------------------------------------------------------------------------------------------------------------

class Crypt(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def encrypt(key, plain_text):
        key_bytes = key.encode()
        source = plain_text.encode()

        long_key = SHA256.new(key_bytes).digest()           # use SHA-256 over our key to get a proper-sized AES key
        iv = Random.new().read(AES.block_size)              # generate IV
        encryptor = AES.new(long_key, AES.MODE_CBC, iv)

        padding = AES.block_size - len(source) % AES.block_size     # calculate required padding
        source += bytes([padding]) * padding
        data = iv + encryptor.encrypt(source)               # store the IV at the beginning and encrypt

        return b64encode(data).decode()


    @staticmethod
    def decrypt(key, cypher_text):
        key_bytes = key.encode()
        source = b64decode(cypher_text.encode())

        long_key = SHA256.new(key_bytes).digest()           # use SHA-256 over our key to get a proper-sized AES key
        iv = source[:AES.block_size]                        # extract the IV from the beginning
        decrypter = AES.new(long_key, AES.MODE_CBC, iv)

        data = decrypter.decrypt(source[AES.block_size:])   # decrypt
        padding = data[-1]                                  # extract the padding value from the end

        if data[-padding:] != bytes([padding]) * padding:
            raise KeyError(key)

        return data[:-padding].decode()                     # remove the padding



"""
Created on 30 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Latitude and longitude, with time of position fix and status
$xxGLL,lat,NS,long,EW,time,status,posMode*cs

example sentence:
$GPGLL,5049.37823,N,00007.37872,W,103228.00,A,D*7F

example values:
GPGLL:{loc:GPLoc:{lat:5049.38432, ns:N, lng:00007.37801, ew:W}, time:GPTime:{time:141058.00}, status:A, pos_mode:D}
GPGLL:{loc:GPLoc:{lat:None, ns:None, lng:None, ew:None}, time:GPTime:{time:140047.00}, status:V, pos_mode:N}

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183
"""

from scs_core.position.nmea.gploc import GPLoc
from scs_core.position.nmea.gptime import GPTime
from scs_core.position.nmea.nmea_sentence import NMEASentence


# --------------------------------------------------------------------------------------------------------------------

class GPGLL(NMEASentence):
    """
    classdocs
    """

    MESSAGE_IDS = ("$GNGLL", "$GPGLL")

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, r):
        if r.message_id not in cls.MESSAGE_IDS:
            raise TypeError("invalid sentence:%s" % r)

        lat = r.str(1)
        ns = r.str(2)

        lng = r.str(3)
        ew = r.str(4)

        loc = GPLoc(lat, ns, lng, ew)

        time = GPTime(r.str(5))

        status = r.str(6)
        pos_mode = r.str(7)

        return GPGLL(r.message_id, loc, time, status, pos_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, message_id, loc, time, status, pos_mode):
        """
        Constructor
        """
        super().__init__(message_id)

        self.__loc = loc                    # GPLoc
        self.__time = time                  # GPTime

        self.__status = status              # string
        self.__pos_mode = pos_mode          # string


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def loc(self):
        return self.__loc


    @property
    def time(self):
        return self.__time


    @property
    def status(self):
        return self.__status


    @property
    def pos_mode(self):
        return self.__pos_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPGLL:{source:%s, loc:%s, time:%s, status:%s, pos_mode:%s}" % \
               (self.source, self.loc, self.time, self.status, self.pos_mode)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"encoding": "utf-8", "content-type": "application/json",
    "text": "{\"rec\": \"2016-11-19T20:31:23.882+00:00\", \"val\": {\"host\": {\"tmp\": 46.2}}}"}}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class MessagePayload(JSONable):
    """
    classdocs
   """

    __TYPE_JSON =       'application/json'
    __TYPE_TEXT =       'text/plain'


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        encoding = jdict.get('encoding')
        content_type = jdict.get('content-type')
        text = jdict.get('text')

        content = PathDict.construct_from_jstr(text) if content_type == cls.__TYPE_JSON else text

        return MessagePayload(encoding, content_type, content)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, encoding, content_type, content):
        """
        Constructor
        """
        self.__encoding = encoding              # string			utf-8
        self.__content_type = content_type      # string			application/json

        self.__content = content                # string			JSON document


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['encoding'] = self.encoding
        jdict['content-type'] = self.content_type

        jdict['content'] = self.content

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def encoding(self):
        return self.__encoding


    @property
    def content_type(self):
        return self.__content_type


    @property
    def content(self):
        return self.__content


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessagePayload:{encoding:%s, content_type:%s, content:%s}" % \
               (self.encoding, self.content_type, self.content)



"""
Created on 17 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"user_id": "southcoastscience-dev", "client-id": "5401", "client-password": "wtxSrs2e"}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class ClientAuth(PersistentJSONable):
    """
    classdocs
    """

    MQTT_HOST =        "mqtt.opensensors.io"          # hard-coded URL

    MQTT_TIMEOUT =     20.0         # seconds


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "osio_client_auth.json"

    @classmethod
    def persistence_location(cls):
        return cls.osio_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        user_id = jdict.get('user-id')

        client_id = jdict.get('client-id')
        client_password = jdict.get('client-password')

        return ClientAuth(user_id, client_id, client_password)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, user_id, client_id, client_password):
        """
        Constructor
        """
        super().__init__()

        self.__user_id = user_id                        # string

        self.__client_id = client_id                    # string
        self.__client_password = client_password        # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['user-id'] = self.user_id

        jdict['client-id'] = self.client_id
        jdict['client-password'] = self.client_password

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def user_id(self):
        return self.__user_id


    @property
    def client_id(self):
        return self.__client_id


    @property
    def client_password(self):
        return self.__client_password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ClientAuth:{user_id:%s, client_id:%s, client_password:%s}" % \
                    (self.user_id, self.client_id, self.client_password)



"""
Created on 8 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"peers": {"scs-bbe-002": {"hostname": "scs-bbe-002", "tag": "scs-be2-2", "shared-secret": "secret1",
"topic": "south-coast-science-dev/production-test/device/alpha-bb-eng-000002/control"}}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable, PersistentJSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class MQTTPeer(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        hostname = jdict.get('hostname')
        tag = jdict.get('tag')
        shared_secret = jdict.get('shared-secret')
        topic = jdict.get('topic')

        return MQTTPeer(hostname, tag, shared_secret, topic)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, hostname, tag, shared_secret, topic):
        """
        Constructor
        """
        self.__hostname = hostname                              # string
        self.__tag = tag                                        # string
        self.__shared_secret = shared_secret                    # string
        self.__topic = topic                                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['hostname'] = self.hostname
        jdict['tag'] = self.tag
        jdict['shared-secret'] = self.shared_secret
        jdict['topic'] = self.topic

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def hostname(self):
        return self.__hostname


    @property
    def tag(self):
        return self.__tag


    @property
    def shared_secret(self):
        return self.__shared_secret


    @shared_secret.setter
    def shared_secret(self, shared_secret):
        self.__shared_secret = shared_secret


    @property
    def topic(self):
        return self.__topic


    @topic.setter
    def topic(self, topic):
        self.__topic = topic


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTPeer:{hostname:%s, tag:%s, shared_secret:%s, topic:%s}" % \
               (self.hostname, self.tag, self.shared_secret, self.topic)


# --------------------------------------------------------------------------------------------------------------------

class MQTTPeerSet(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME =    "mqtt_peers.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    @classmethod
    def load(cls, manager, encryption_key=None, skeleton=False):
        instance = super().load(manager, encryption_key=encryption_key, skeleton=skeleton)

        if instance is None:
            instance = cls(OrderedDict())

        return instance


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        peers = OrderedDict()

        for hostname, item in jdict.get('peers').items():
            peer = MQTTPeer.construct_from_jdict(item)

            peers[hostname] = peer

        return cls(peers)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, peers):
        """
        Constructor
        """
        super().__init__()

        self.__peers = peers                                # OrderedDict of string: MQTTPeer


    def __len__(self):
        return len(self.__peers)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['peers'] = self.__peers

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def insert(self, peer: MQTTPeer):
        # add...
        self.__peers[peer.hostname] = peer

        # sort...
        peers = OrderedDict()

        for hostname in sorted(self.__peers.keys()):
            peers[hostname] = self.__peers[hostname]

        self.__peers = peers


    def remove(self, hostname):
        try:
            del(self.__peers[hostname])
            return True

        except KeyError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def subset(self, hostname_substring=None, topic_substring=None):
        subset = OrderedDict()

        for hostname, peer in self.__peers.items():
            if hostname_substring is not None and hostname_substring not in hostname:
                continue

            if topic_substring is not None and topic_substring not in peer.topic:
                continue

            subset[peer.hostname] = peer

        return MQTTPeerSet(subset)


    def peer(self, hostname):
        try:
            return self.__peers[hostname]
        except KeyError:
            return None


    def peer_by_tag(self, tag):
        for peer in self.__peers.values():
            if peer.tag == tag:
                return peer

        return None


    @property
    def peers(self):
        return list(self.__peers.values())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MQTTPeerSet:{peers:%s}" % Str.collection(self.__peers)



"""
Created on 15 Oct 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from scs_core.comms.uds_client import UDSClient

from scs_core.data.json import JSONify
from scs_core.data.linear_regression import LinearRegression
from scs_core.data.path_dict import PathDict

from scs_core.model.gas.gas_inference_client import GasInferenceClient
from scs_core.model.gas.vE.gas_request import GasRequest

from scs_core.sync.schedule import ScheduleItem

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class VEGasInferenceClient(GasInferenceClient):
    """
    classdocs
    """

    @classmethod
    def construct(cls, socket, inference_uds_path, schedule_item: ScheduleItem, vcal_baseline, gas_baseline,
                  model_compendium_group):
        # UDS...
        uds_client = UDSClient(socket, inference_uds_path)

        # T / rH slope...
        slope_tally = GasRequest.slope_tally(schedule_item.duration())

        t_regression = LinearRegression(tally=slope_tally)
        rh_regression = LinearRegression(tally=slope_tally)

        return cls(uds_client, t_regression, rh_regression, vcal_baseline, gas_baseline, model_compendium_group)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_client, t_regression, rh_regression, vcal_baseline, gas_baseline, model_compendium_group):
        """
        Constructor
        """
        super().__init__(uds_client)

        self.__t_regression = t_regression                              # LinearRegression
        self.__rh_regression = rh_regression                            # LinearRegression

        self.__vcal_baseline = vcal_baseline                            # VCalBaseline
        self.__gas_baseline = gas_baseline                              # GasBaseline
        self.__model_compendium_group = model_compendium_group          # ModelCompendiumGroup

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def infer(self, gas_sample, board_temp):
        # T / rH slope...
        self.__t_regression.append(gas_sample.rec, gas_sample.sht_datum.temp)
        self.__rh_regression.append(gas_sample.rec, gas_sample.sht_datum.humid)

        m, _ = self.__t_regression.line()
        t_slope = 0.0 if m is None else m

        m, _ = self.__rh_regression.line()
        rh_slope = 0.0 if m is None else m

        # preprocess...
        request = PathDict(GasRequest(gas_sample, t_slope, rh_slope, board_temp).as_json())
        sample = request.node(sub_path='sample')

        preprocessed = self.__model_compendium_group.preprocess(request, self.__vcal_baseline)

        # infer...
        self._uds_client.request(JSONify.dumps(preprocessed))
        response = PathDict(json.loads(self._uds_client.wait_for_response()))

        if not response:
            self.__logger.error("request rejected: %s" % JSONify.dumps(gas_sample))
            return sample

        # postprocess...
        exg = PathDict(self.__model_compendium_group.postprocess(preprocessed, response))

        # gas baseline...
        for gas, offset in self.__gas_baseline.offsets().items():
            path = '.'.join(('val', gas, 'cnc'))
            baselined_cnc = round(float(exg.node(sub_path=path)) + offset, 1)
            exg.append(path, round(baselined_cnc, 1))

        # report...
        report = PathDict(sample)                       # discard any changes made in preprocessing

        if exg is not None:
            report.append('ver', response.node(sub_path='ver'))
            report.append('exg', exg.node())

        return report.dictionary


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "VEGasInferenceClient:{uds_client:%s, t_regression:%s, rh_regression:%s, vcal_baseline:%s, " \
               "gas_baseline:%s, model_compendium_group:%s}" %  \
               (self._uds_client, self.__t_regression, self.__rh_regression, self.__vcal_baseline,
                self.__gas_baseline, self.__model_compendium_group)



"""
Created on 26 Jun 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import signal
import sys


# --------------------------------------------------------------------------------------------------------------------

class SignalledExit(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, client, verbose):
        listener = cls(client, verbose)
        listener.set()

        return listener


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client, verbose):
        """
        Constructor
        """
        self.__client = client                                  # string
        self.__verbose = verbose                                # bool

        self.__original_sigint_handler = None                   # function?


    # ----------------------------------------------------------------------------------------------------------------

    def sigint_handler(self, signum, _frame):
        self.clear()

        if self.__verbose:
            print("%s: SIGINT (%d)" % (self.__client, signum), file=sys.stderr)
            sys.stderr.flush()

        sys.exit(1)


    def sigterm_handler(self, signum, _frame):
        self.clear()

        if self.__verbose:
            print("%s: SIGTERM (%d)" % (self.__client, signum), file=sys.stderr)
            sys.stderr.flush()

        sys.exit(0)


    # ----------------------------------------------------------------------------------------------------------------

    def set(self):
        if self.__original_sigint_handler is None:
            self.__original_sigint_handler = signal.getsignal(signal.SIGINT)

        signal.signal(signal.SIGINT, self.sigint_handler)
        signal.signal(signal.SIGTERM, self.sigterm_handler)


    def clear(self):
        if self.__original_sigint_handler is None:
            return

        signal.signal(signal.SIGINT, self.__original_sigint_handler)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SignalledExit:{client:%s, verbose:%s, original_sigint_handler:%s}" % \
               (self.__client, self.__verbose, self.__original_sigint_handler)



"""
Created on 13 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"root-path": "/home/pi/SCS/logs", "delete-oldest": true, "write-interval": 0}
"""

from collections import OrderedDict

from scs_core.csv.csv_log import CSVLog
from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class CSVLoggerConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "csv_logger_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        root_path = jdict.get('root-path')
        delete_oldest = jdict.get('delete-oldest')
        write_interval = jdict.get('write-interval')

        return CSVLoggerConf(root_path, delete_oldest, write_interval)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, root_path, delete_oldest, write_interval):
        """
        Constructor
        """
        super().__init__()

        self.__root_path = root_path                            # string
        self.__delete_oldest = bool(delete_oldest)              # bool
        self.__write_interval = int(write_interval)             # int


    def __eq__(self, other):
        try:
            return self.root_path == other.root_path and self.delete_oldest == other.delete_oldest and \
                   self.write_interval == other.write_interval

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def csv_log(self, topic_subject, tag=None, timeline_start=None):
        return CSVLog(self.root_path, topic_subject, tag=tag, timeline_start=timeline_start)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['root-path'] = self.root_path
        jdict['delete-oldest'] = self.delete_oldest
        jdict['write-interval'] = self.write_interval

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def root_path(self):
        return self.__root_path


    @property
    def delete_oldest(self):
        return self.__delete_oldest


    @property
    def write_interval(self):
        return self.__write_interval


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVLoggerConf:{root_path:%s, delete_oldest:%s, write_interval:%s}" %  \
               (self.root_path, self.delete_oldest, self.write_interval)



"""
Created on 14 Aug 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import time

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class UDSClient(object):
    """
    classdocs
    """

    EOS = 'UDS-CLIENT-EOS'                          # end of session message

    _RECONNECT_WAIT = 10.0                          # seconds

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, socket, path):
        """
        Constructor
        """
        self.__logger = Logging.getLogger()

        self.__path = path                          # string

        self.__uds = socket(path)                   # DomainSocket
        self.__disconnecting = False                # bool


    # ----------------------------------------------------------------------------------------------------------------

    def request(self, message):
        reported_waiting = False

        while True:
            try:
                self.__uds.client_send(message.strip())
                return

            except (BrokenPipeError, OSError):
                if self.__disconnecting:
                    return                          # don't wait forever if terminating

                if not reported_waiting:
                    self.__logger.info('waiting for server')
                    reported_waiting = True

                time.sleep(self._RECONNECT_WAIT)

                self.__uds.close()                  # attempt to restart session
                self.__uds.connect()


    def wait_for_response(self):
        return self.__uds.client_receive()


    # ----------------------------------------------------------------------------------------------------------------

    def open(self):
        self.__disconnecting = False
        self.__uds.connect()

        self.__logger.info('opened UDS')


    def close(self):
        self.__disconnecting = True
        self.request(self.EOS)
        self.__uds.close()

        self.__logger.info('closed UDS')


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "UDSClient:{uds:%s}" % self.__uds



"""
Created on 20 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
event: message
data: {"topic":"/users/southcoastscience-dev/test/json","date":1479655521714,
            "message":"{\"rec\": \"2016-11-20T15:25:29.605+00:00\", \"val\": {\"host\": {\"tmp\": 46.2}}}"}
"""

import json

from scs_core.osio.client.realtime_client import RealtimeClient
from scs_core.osio.data.message_event import MessageEvent


# --------------------------------------------------------------------------------------------------------------------

class MessageEventSubscriber(object):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, streaming_client):
        self.__streaming_client = streaming_client

        self.__listener = None
        self.__realtime_client = None


    # ----------------------------------------------------------------------------------------------------------------

    def subscribe(self, listener, auth, topic):
        self.close()

        self.__listener = listener

        self.__realtime_client = RealtimeClient(self.__streaming_client, self.__local_listener, auth)
        self.__realtime_client.connect(topic)


    def close(self):                        # TODO: sort out order of closing
        if not self.__realtime_client:
            return

        self.__realtime_client.close()
        self.__realtime_client = None


    # ----------------------------------------------------------------------------------------------------------------

    def __local_listener(self, datum):
        if not self.__listener:
            return

        jdict = json.loads(datum)
        event = MessageEvent.construct_from_jdict(jdict)

        self.__listener(event)


# ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageEventSubscriber:{streaming_client:%s}" % self.__streaming_client



"""
Created on 2 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
  "description": "electrochemical we (V), ae (V), wc (V), cnc (ppb), Pt100 temp, internal SHT",
  "unit": null,
  "derived-topics": [
    {
      "stats": "electrochemical we (V), ae (V), wc (V), cnc (ppb), Pt100 temp, internal SHT",
      "unit": null,
      "name": "hourly statistics of Gas concentrations",
      "public": true,
      "topic": "/osio/orgs/south-coast-science-dev/production-test/loc/1/gases/hourly",
      "derived-data": {
        "interval": 3600
      },
      "topic-info": {
        "format": "application/json"
      }
    }
  ],
  "name": "Gas concentrations",
  "public": true,
  "topic": "/orgs/south-coast-science-dev/production-test/loc/1/gases",
  "topic-info": {
    "format": "application/json"
  },
  "bookmark-count": 0,
  "stats": {
    "period": 604800,
    "last-reading": "5 seconds ago",
    "average-frequency": "9 seconds",
    "total": 49661,
    "contributors": [
      {
        "name": "South Coast Science - Dev",
        "id": "southcoastscience-dev",
        "gravatar-hash": "07f512e9fe64863039df0c0f1834cc25"
      }
    ],
    "last-location": {
      "lat": 50.819456,
      "lon": -0.128336
    }
  }
}
"""

from scs_core.data.str import Str

from scs_core.osio.data.abstract_topic import AbstractTopic
from scs_core.osio.data.derived_topic import DerivedTopic
from scs_core.osio.data.topic_info import TopicInfo
from scs_core.osio.data.topic_stats import TopicStats


# --------------------------------------------------------------------------------------------------------------------

class TopicMetadata(AbstractTopic):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # AbstractTopic...
        path = jdict.get('topic')
        name = jdict.get('name')
        description = jdict.get('description')

        is_public = jdict.get('public')

        info = TopicInfo.construct_from_jdict(jdict.get('topic-info'))

        # TopicMetadata...
        if jdict.get('derived-topics'):
            derived_topics = [DerivedTopic.construct_from_jdict(dt_jdict) for dt_jdict in jdict.get('derived-topics')]
        else:
            derived_topics = []

        bookmark_count = jdict.get('bookmark-count')
        stats = TopicStats.construct_from_jdict(jdict.get('stats'))

        return TopicMetadata(path, name, description, is_public, info, derived_topics, bookmark_count, stats)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, name, description, is_public, info,
                 derived_topics, bookmark_count, stats):
        """
        Constructor
        """

        # AbstractTopic...
        AbstractTopic.__init__(self, path, name, description, is_public, info)

        # TopicMetadata...
        self.__derived_topics = derived_topics              # list of DerivedTopic
        self.__bookmark_count = bookmark_count              # int
        self.__stats = stats                                # TopicStats


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = AbstractTopic.as_json(self)

        jdict['derived-topics'] = self.derived_topics
        jdict['bookmark-count'] = self.bookmark_count
        jdict['stats'] = self.stats

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def derived_topics(self):
        return self.__derived_topics


    @property
    def bookmark_count(self):
        return self.__bookmark_count


    @property
    def stats(self):
        return self.__stats


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicMetadata:{path:%s, name:%s, description:%s, is_public:%s, info:%s, " \
               "derived_topics:%s, bookmark_count:%s, stats:%s}" % \
               (self.path, self.name, self.description, self.is_public, self.info,
                Str.collection(self.derived_topics), self.bookmark_count, self.stats)



"""
Created on 13 Jun 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

command example:
ps -Ao ppid,pid,uid,tty,pmem,pcpu,cputime,etime,args -w -w

report examples:
  154 26605   502 ??        0.0   0.0   0:00.37 01-00:43:12 /usr/bin/ssh-agent -l
  154 39364   502 ??        6.3  31.5  22:58.40    02:26:25 /Applications/PyCharm.app/Contents/MacOS/pycharm

JSON example:
{"ppid": 154, "pid": 26605, "uid": 502, "tty": "??", "pcpu": 0.0, "pmem": 0.0,
"cpu": {"days": 0, "hours": 0, "minutes": 0, "seconds": 0, "millis": 370},
"elapsed": {"days": 1, "hours": 0, "minutes": 43, "seconds": 12, "millis": 0},
"cmd": "/usr/bin/ssh-agent -l"}
"""

import re

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class PsDatum(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_report(cls, report):
        mat = re.match(r'^\s*(\d+)\s+(\d+)\s+(\d+)\s+([^\s]+)\s+(\d+\.\d+)\s+(\d+\.\d+)\s+([^\s]+)\s+([^\s]+)\s+(.+)$',
                       report)

        if mat is None:
            return None

        fields = mat.groups()

        ppid = int(fields[0])
        pid = int(fields[1])
        uid = int(fields[2])
        tty = fields[3]

        pcpu = float(fields[4])
        pmem = float(fields[5])

        cpu_time = Timedelta.construct_from_ps_time_report(fields[6])
        elapsed_time = Timedelta.construct_from_ps_elapsed_report(fields[7])
        cmd = fields[8]

        return PsDatum(ppid, pid, uid, tty, pcpu, pmem, cpu_time, elapsed_time, cmd)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        ppid = int(jdict.get('ppid'))
        pid = int(jdict.get('pid'))
        uid = int(jdict.get('uid'))
        tty = jdict.get('tty')

        pcpu = float(jdict.get('pcpu'))
        pmem = float(jdict.get('pmem'))

        cpu_time = Timedelta.construct_from_jdict(jdict.get('cpu'))
        elapsed_time = Timedelta.construct_from_jdict(jdict.get('elapsed'))

        cmd = jdict.get('cmd')

        return PsDatum(ppid, pid, uid, tty, pcpu, pmem, cpu_time, elapsed_time, cmd)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, ppid, pid, uid, tty, pcpu, pmem, cpu_time, elapsed_time, cmd):
        """
        Constructor
        """
        self.__ppid = ppid                  # int               parent process ID
        self.__pid = pid                    # int               process ID
        self.__uid = uid                    # int               user ID
        self.__tty = tty                    # string            TTY

        self.__pcpu = pcpu                  # float             % CPU
        self.__pmem = pmem                  # float             % memory

        self.__cpu_time = cpu_time          # Timedelta
        self.__elapsed_time = elapsed_time  # Timedelta

        self.__cmd = cmd                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['ppid'] = self.ppid
        jdict['pid'] = self.pid
        jdict['uid'] = self.uid
        jdict['tty'] = self.tty

        jdict['pcpu'] = self.pcpu
        jdict['pmem'] = self.pmem

        jdict['cpu'] = self.cpu_time
        jdict['elapsed'] = self.elapsed_time

        jdict['cmd'] = self.cmd

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def ppid(self):
        return self.__ppid


    @property
    def pid(self):
        return self.__pid


    @property
    def uid(self):
        return self.__uid


    @property
    def tty(self):
        return self.__tty


    @property
    def pcpu(self):
        return self.__pcpu


    @property
    def pmem(self):
        return self.__pmem


    @property
    def cpu_time(self):
        return self.__cpu_time


    @property
    def elapsed_time(self):
        return self.__elapsed_time


    @property
    def cmd(self):
        return self.__cmd


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PsDatum:{ppid:%s, pid:%s, uid:%s, tty:%s, pcpu:%s, pmem:%s, " \
               "cpu_time:%s, elapsed_time:%s, cmd:%s}" %  \
               (self.ppid, self.pid, self.uid, self.tty, self.pcpu, self.pmem,
                self.cpu_time, self.elapsed_time, self.cmd)



"""
Created on 21 Jan 2021

@author: Jade Page (jade.page@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Deployment(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_aws(cls, group_name, jdict):
        if not group_name:
            return None

        if not jdict:
            return cls(group_name, None, None, None)

        group_name = group_name
        created_at = LocalizedDatetime.construct_from_iso8601(jdict.get('CreatedAt'))
        deployment_type = jdict.get("DeploymentType")
        deployment_id = jdict.get("DeploymentId")

        return cls(group_name, created_at, deployment_type, deployment_id)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        group_name = jdict.get('group_name')
        created_at = LocalizedDatetime.construct_from_iso8601(jdict.get('created_at'))
        deployment_type = jdict.get('deployment_type')
        deployment_id = jdict.get('deployment_id')

        return cls(group_name, created_at, deployment_type, deployment_id)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, group_name, created_at, deployment_type, deployment_id):
        """
        Constructor
        """
        self.__group_name = group_name                      # string
        self.__created_at = created_at                      # LocalizedDatetime
        self.__deployment_type = deployment_type            # string
        self.__deployment_id = deployment_id                # string


    def __lt__(self, other):
        return self.group_name < other.group_name


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['group_name'] = self.group_name
        jdict['created_at'] = None if self.created_at is None else self.created_at.as_iso8601()
        jdict['deployment_type'] = self.deployment_type
        jdict['deployment_id'] = self.deployment_id

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def is_current(self, datetime):
        if self.created_at is None:
            return False

        return self.created_at >= datetime


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def group_name(self):
        return self.__group_name


    @property
    def created_at(self):
        return self.__created_at


    @property
    def deployment_type(self):
        return self.__deployment_type


    @property
    def deployment_id(self):
        return self.__deployment_id


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Deployment:{group_name:%s, created_at:%s, deployment_type:%s, deployment_id:%s}" % \
               (self.group_name, self.created_at, self.deployment_type, self.deployment_id)



"""
Created on 8 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import urllib.parse

from scs_core.osio.client.rest_client import RESTClient
from scs_core.osio.data.organisation import Organisation


# --------------------------------------------------------------------------------------------------------------------

class OrganisationManager(object):
    """
    classdocs
    """
    __FINDER_BATCH_SIZE = 100

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, org_id):
        request_path = '/v1/orgs/' + urllib.parse.quote(org_id, '')

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.get(request_path)
        except RuntimeError:
            response_jdict = None

        self.__rest_client.close()

        topic = Organisation.construct_from_jdict(response_jdict)

        return topic


    def find_owned_by_user(self, org_id):
        orgs = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__get(org_id):
                orgs.extend(batch)

        finally:
            self.__rest_client.close()

        return orgs


    # ----------------------------------------------------------------------------------------------------------------

    def create(self, org):
        request_path = '/v1/orgs'

        # request...
        self.__rest_client.connect()

        try:
            response = self.__rest_client.post(request_path, org.as_json())

        finally:
            self.__rest_client.close()

        print("create response: %s" % response)

        return response


    def update(self, org_id, org):
        request_path = '/v1/orgs/' + org_id

        # request...
        self.__rest_client.connect()

        try:
            self.__rest_client.put(request_path, org.as_json())
        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __get(self, user_id):
        request_path = '/v12/users/' + user_id + '/owned-orgs'
        params = {'offset': 0, 'count': self.__FINDER_BATCH_SIZE}

        while True:
            # request...
            response_jdict = self.__rest_client.get(request_path, params=params)

            # organisations...
            orgs = [Organisation.construct_from_jdict(org_jdict) for org_jdict in response_jdict] \
                if response_jdict else []

            yield orgs

            if len(orgs) == 0:
                break

            # next...
            params['offset'] += len(orgs)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationManager:{rest_client:%s}" % self.__rest_client



"""
Created on 25 Jul 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://howchoo.com/g/ywi5m2vkodk/working-with-datetime-objects-and-timezones-in-python

example JSON:
{"timezone": "Europe/London", "start-hour": 17, "end-hour": 8, "aggregation-period": {"interval": 5, "units": "M"},
"gas-minimums": {"CO": 200, "CO2": 420, "H2S": 5, "NO": 10, "NO2": 10, "SO2": 5}}
"""

import pytz

from collections import OrderedDict
from datetime import datetime

from scs_core.data.json import MultiPersistentJSONable
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.recurring_period import RecurringPeriod
from scs_core.data.str import Str
from scs_core.data.timedelta import Timedelta


# --------------------------------------------------------------------------------------------------------------------

class BaselineConf(MultiPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __SUPPORTED_GASES = ('CO', 'CO2', 'H2S', 'NO', 'NO2', 'Ox', 'SO2', 'VOC', 'VOCe')

    @classmethod
    def supported_gases(cls):
        return cls.__SUPPORTED_GASES


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "baseline_conf.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.conf_dir(), filename


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return None

        timezone = jdict.get('timezone')
        start_hour = jdict.get('start-hour')
        end_hour = jdict.get('end-hour')
        aggregation_period = RecurringPeriod.construct_from_jdict(jdict.get('aggregation-period'))
        minimums = jdict.get('minimums')

        return cls(name, timezone, start_hour, end_hour, aggregation_period, minimums)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, timezone, start_hour, end_hour, aggregation_period, minimums):
        """
        Constructor
        """
        super().__init__(name)

        self.__timezone = timezone                              # string
        self.__start_hour = int(start_hour)                     # int
        self.__end_hour = int(end_hour)                         # int
        self.__aggregation_period = aggregation_period          # RecurringMinutes
        self.__minimums = minimums                              # dict of string: int


    # ----------------------------------------------------------------------------------------------------------------

    def start_datetime(self, origin: LocalizedDatetime):
        dt = origin.datetime
        tz = pytz.timezone(self.timezone)

        start_dt = datetime(dt.year, month=dt.month, day=dt.day, hour=self.start_hour)
        start = LocalizedDatetime(tz.localize(start_dt))

        if self.start_hour > self.end_hour:
            start -= Timedelta(days=1)

        return start.utc()


    def end_datetime(self, origin: LocalizedDatetime):
        dt = origin.datetime
        tz = pytz.timezone(self.timezone)

        end_dt = datetime(dt.year, month=dt.month, day=dt.day, hour=self.end_hour)
        end = LocalizedDatetime(tz.localize(end_dt))

        return end.utc()


    def expected_data_points(self, start, end):
        points_per_hour = 60 / self.aggregation_period.interval
        period = end - start

        return int(points_per_hour * period.hours)


    def checkpoint(self):
        return self.aggregation_period.checkpoint()


    # ----------------------------------------------------------------------------------------------------------------

    def set_minimum(self, gas, value):
        self.__minimums[gas] = value


    def remove_minimum(self, gas):
        try:
            del self.__minimums[gas]
        except KeyError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def timezone(self):
        return self.__timezone


    @property
    def start_hour(self):
        return self.__start_hour


    @property
    def end_hour(self):
        return self.__end_hour


    @property
    def aggregation_period(self):
        return self.__aggregation_period


    def minimum(self, gas):
        return self.__minimums[gas]


    @property
    def minimums(self):
        return OrderedDict(sorted(self.__minimums.items()))


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['timezone'] = self.timezone
        jdict['start-hour'] = self.start_hour
        jdict['end-hour'] = self.end_hour
        jdict['aggregation-period'] = self.aggregation_period.as_json()
        jdict['minimums'] = self.minimums

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        minimums = Str.collection(self.minimums)

        return "BaselineConf:{name:%s, timezone:%s, start_hour:%s, end_hour:%s, aggregation_period:%s, " \
               "minimums:%s}" %  \
               (self.name, self.timezone, self.start_hour, self.end_hour, self.aggregation_period,
                minimums)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
  "description": "BB development platform",
  "tags": [
    "pm1",
    "pm2.5",
    "pm10",
    "co",
    "no",
    "no2",
    "o3",
    "temperature",
    "humidity"
  ],
  "client-id": "5926",
  "name": "Alpha Pi Eng/V1 000006",
  "org-id": "south-coast-science-dev",
  "batch": null,
  "device-type": "Alpha Pi Eng/V1",
  "owner-id": "southcoastscience-dev",
  "location": {
    "lat": 50.819456,
    "lon": -0.128336,
    "postcode": "bn2 1af"
  }
}
"""

from scs_core.osio.data.device_summary import DeviceSummary
from scs_core.osio.data.location import Location


# --------------------------------------------------------------------------------------------------------------------

class Device(DeviceSummary):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # DeviceSummary...
        client_id = jdict.get('client-id')
        name = jdict.get('name')
        description = jdict.get('description')

        location = Location.construct_from_jdict(jdict.get('location'))

        tags = jdict.get('tags')

        # Device...
        password = jdict.get('password')
        password_is_locked = jdict.get('password-is-locked')

        device_type = jdict.get('device-type')
        batch = jdict.get('batch')

        org_id = jdict.get('org-id')
        owner_id = jdict.get('owner-id')

        device = Device(client_id, name, description, password, password_is_locked, location,
                        device_type, batch, org_id, owner_id, tags)

        return device


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client_id, name, description, password, password_is_locked, location,
                 device_type, batch, org_id, owner_id, tags):
        """
        Constructor
        """
        DeviceSummary.__init__(self, client_id, name, description, location, tags)

        self.__password = password                          # string
        self.__password_is_locked = password_is_locked      # bool

        self.__device_type = device_type                    # string
        self.__batch = batch                                # string

        self.__org_id = org_id                              # string
        self.__owner_id = owner_id                          # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = DeviceSummary.as_json(self)

        if self.password is not None:
            jdict['password'] = self.password

        if self.password_is_locked is not None:
            jdict['password-is-locked'] = self.password_is_locked

        jdict['device-type'] = self.device_type

        if self.batch is not None:
            jdict['batch'] = self.batch

        if self.org_id is not None:
            jdict['org-id'] = self.org_id

        if self.owner_id is not None:
            jdict['owner-id'] = self.owner_id

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def password(self):
        return self.__password


    @property
    def password_is_locked(self):
        return self.__password_is_locked


    @property
    def device_type(self):
        return self.__device_type


    @property
    def batch(self):
        return self.__batch


    @property
    def org_id(self):
        return self.__org_id


    @property
    def owner_id(self):
        return self.__owner_id


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Device:{client_id:%s, name:%s, description:%s, password:%s, password_is_locked:%s, " \
               "location:%s, device_type:%s, batch:%s, org_id:%s, owner_id:%s, tags:%s}" % \
                    (self.client_id, self.name, self.description, self.password, self.password_is_locked,
                     self.location, self.device_type, self.batch, self.org_id, self.owner_id, self.tags)



"""
Created on 6 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
    {
      "name": "Particulate densities",
      "description": "pm1 (ug/m3), pm2.5 (ug/m3), pm10 (ug/m3), bin counts, mtf1, mtf3, mtf5 mtf7",
      "topic": "/orgs/south-coast-science-dev/development/loc/1/particulates",
      "public": true,
      "schema": {
        "id": 29,
        "name": "south-coast-science-particulates"
      },
      "rollups-enabled": true,
      "topic-info": {
        "format": "application/json"
      }
    }
"""

from scs_core.osio.data.abstract_topic import AbstractTopic
from scs_core.osio.data.schema import Schema
from scs_core.osio.data.topic_info import TopicInfo


# --------------------------------------------------------------------------------------------------------------------

class TopicSummary(AbstractTopic):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # AbstractTopic...
        path = jdict.get('topic')
        name = jdict.get('name')
        description = jdict.get('description')

        is_public = jdict.get('public')

        info = TopicInfo.construct_from_jdict(jdict.get('topic-info'))

        # TopicSummary...
        rollups_enabled = jdict.get('rollups-enabled')
        schema = Schema.construct_from_jdict(jdict.get('schema'))

        return TopicSummary(path, name, description, is_public, info, rollups_enabled, schema)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, name, description, is_public, info,
                 rollups_enabled, schema):
        """
        Constructor
        """
        # AbstractTopic...
        AbstractTopic.__init__(self, path, name, description, is_public, info)

        # TopicSummary...
        self.__rollups_enabled = rollups_enabled        # bool
        self.__schema = schema                          # Schema


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = AbstractTopic.as_json(self)

        jdict['rollups-enabled'] = self.rollups_enabled

        if self.schema is not None:
            jdict['schema'] = self.schema

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def schema_id(self):
        if self.__schema is None:
            return None

        return self.__schema.id


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def rollups_enabled(self):
        return self.__rollups_enabled


    @property
    def schema(self):
        return self.__schema


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicSummary:{path:%s, name:%s, description:%s, is_public:%s, info:%s, " \
               "rollups_enabled:%s, schema:%s}" % \
               (self.path, self.name, self.description, self.is_public, self.info,
                self.rollups_enabled, self.schema)



"""
Created on 24 May 2021

@author: Jade Page (jade.page@southcoastscience.com)

document example:
{"tag": "scs-bgx-003", "rec": "2021-04-14T08:49:22+01:00", "result": "PENDING"}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class GitPullCheck(JSONable):
    """
    classdocs
    """

    RESULT_PENDING =                    "PENDING"
    RESULT_ERROR =                      "ERROR"
    RESULT_DONE =                       "DONE"
    RESULT_NOT_SUPPORTED =              "NOT SUPPORTED"
    RESULT_MALFORMED_GIT_PULL =         "MALFORMED:GIT_PULL"

    __RESULTS = {
        'PEN': RESULT_PENDING,
        'ERR': RESULT_ERROR,
        'NSP': RESULT_NOT_SUPPORTED,
        'MGP': RESULT_MALFORMED_GIT_PULL,
        'D':   RESULT_DONE
    }

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def result_codes(cls):
        return cls.__RESULTS.keys()


    @classmethod
    def result_string(cls, code):
        if code is None:
            return None

        return cls.__RESULTS[code]


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('rec'))
        message_rec = LocalizedDatetime.construct_from_iso8601(jdict.get('message-rec'))
        result = jdict.get('result')
        context = jdict.get('context')

        return cls(tag, rec, message_rec, result, context=context)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, message_rec, result, context=''):
        """
        Constructor
        """
        self.__tag = tag                        # string
        self.__rec = rec                        # LocalizedDatetime - when this record was created
        self.__message_rec = message_rec        # LocalizedDatetime - the rec of the ControlReceipt
        self.__result = result                  # string
        self.__context = context                # string (only for ERROR)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['tag'] = self.tag
        jdict['rec'] = self.rec.as_iso8601()
        jdict['message-rec'] = None if self.message_rec is None else self.message_rec.as_iso8601()
        jdict['result'] = self.result
        jdict['context'] = self.context if self.context else ["-"]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag(self):
        return self.__tag


    @property
    def rec(self):
        return self.__rec


    @property
    def message_rec(self):
        return self.__message_rec


    @property
    def result(self):
        return self.__result

    @property
    def context(self):
        return self.__context


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GitPullCheck:{tag:%s, rec:%s, message_rec:%s, result:%s, context:%s}" % \
               (self.tag, self.rec, self.message_rec, self.result, self.context)



"""
Created on 11 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

settings for OPCMonitor

example JSON:
{"model": "N3", "sample-period": 10, "restart-on-zeroes": true, "power-saving": false}
"""

from collections import OrderedDict

from scs_core.data.json import MultiPersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class OPCConf(MultiPersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "opc_conf.json"

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.conf_dir(), filename


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return None

        model = jdict.get('model')
        sample_period = jdict.get('sample-period')
        restart_on_zeroes = jdict.get('restart-on-zeroes', True)
        power_saving = jdict.get('power-saving')

        bus = jdict.get('bus')
        address = jdict.get('address')

        return cls(model, sample_period, restart_on_zeroes, power_saving, bus, address, name=name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, model, sample_period, restart_on_zeroes, power_saving, bus, address, name=None):
        """
        Constructor
        """
        super().__init__(name)

        self.__model = model                                        # string
        self.__sample_period = int(sample_period)                   # int
        self.__restart_on_zeroes = bool(restart_on_zeroes)          # bool
        self.__power_saving = bool(power_saving)                    # bool

        self.__bus = bus                                            # int
        self.__address = address                                    # int


    def __eq__(self, other):                            # ignore name
        try:
            return self.model == other.model and self.sample_period == other.sample_period and \
                   self.restart_on_zeroes == other.restart_on_zeroes and self.power_saving == other.power_saving and \
                   self.bus == other.bus and self.address == other.address

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def opc_monitor(self, interface, host):
        return None


    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def opc(self, interface, host):
        return None


    # noinspection PyMethodMayBeStatic
    def uses_spi(self):
        return True


    def opc_bus(self, host):
        try:
            return int(self.__bus)

        except TypeError:
            return host.opc_spi_bus()


    def opc_address(self, host):
        try:
            return int(self.__address)

        except TypeError:
            return host.opc_spi_device()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def model(self):
        return self.__model


    @property
    def sample_period(self):
        return self.__sample_period


    @property
    def power_saving(self):
        return self.__power_saving


    @property
    def restart_on_zeroes(self):
        return self.__restart_on_zeroes


    @property
    def bus(self):
        return self.__bus


    @property
    def address(self):
        return self.__address


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['model'] = self.model
        jdict['sample-period'] = self.sample_period
        jdict['restart-on-zeroes'] = self.restart_on_zeroes
        jdict['power-saving'] = self.power_saving

        if self.__bus is not None:
            jdict['bus'] = self.bus

        if self.__address is not None:
            jdict['address'] = self.address

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OPCConf(core):{name:%s, model:%s, sample_period:%s, restart_on_zeroes:%s, power_saving:%s, " \
               "bus:%s, address:%s}" %  \
               (self.name, self.model, self.sample_period, self.restart_on_zeroes, self.power_saving,
                self.bus, self.address)



"""
Created on 9 Jan 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

http://softwareengineering.stackexchange.com/questions/298364/exception-handling-in-python-am-i-doing-this-wrong-and-why
"""

import sys
import traceback

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.str import Str

from scs_core.sys.trace_entry import TraceEntry


# --------------------------------------------------------------------------------------------------------------------

class ExceptionReport(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, ex):
        excls = ex.__class__
        args = ex.args

        info = sys.exc_info()

        if len(info) < 3:
            return ExceptionReport(excls, args, None, None)

        tb = info[2]
        fe = traceback.format_exception(excls, ex, tb)
        entries = [entry.strip() for entry in fe]

        trace = [TraceEntry.construct(line) for line in entries[1:-1]]
        summary = entries[-1]

        return ExceptionReport(excls, args, trace, summary)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, cls, args, trace, summary):
        """
        Constructor
        """
        self.__cls = cls
        self.__args = args
        self.__trace = trace
        self.__summary = summary


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['cls'] = self.__class__.__name__
        jdict['args'] = self.args
        jdict['trace'] = self.trace
        jdict['sum'] = self.summary

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def cls(self):
        return self.__cls


    @property
    def trace(self):
        return self.__trace


    @property
    def summary(self):
        return self.__summary


    @property
    def args(self):
        return self.__args


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ExceptionReport:{cls:%s, args:%s, trace:%s, summary:%s}" % \
               (self.__class__.__name__, Str.collection(self.args), Str.collection(self.trace), self.summary)



"""
Created on 10 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
    {
      "description": "electrochemical we (V), ae (V), wc (V), cnc (ppb), Pt100 temp, internal SHT",
      "unit": null,
      "name": "hourly statistics of Gas concentrations",
      "public": true,
      "topic": "/osio/orgs/south-coast-science-dev/production-test/loc/1/gases/hourly",
      "derived-data": {
        "interval": 3600
      },
      "topic-info": {
        "format": "application/json"
      }
    }
"""

from collections import OrderedDict

from scs_core.osio.data.abstract_topic import AbstractTopic
from scs_core.osio.data.derived_data import DerivedData
from scs_core.osio.data.topic_info import TopicInfo


# --------------------------------------------------------------------------------------------------------------------

class DerivedTopic(AbstractTopic):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # AbstractTopic...
        path = jdict.get('topic')
        name = jdict.get('name')
        description = jdict.get('description')

        is_public = jdict.get('public')

        info = TopicInfo.construct_from_jdict(jdict.get('topic-info'))

        # DerivedTopic...
        unit = jdict.get('unit')
        derived_data = DerivedData.construct_from_jdict(jdict.get('derived-data'))

        return DerivedTopic(path, name, description, is_public, info, unit, derived_data)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, name, description, is_public, info,
                 unit, derived_data):
        """
        Constructor
        """
        # AbstractTopic...
        AbstractTopic.__init__(self, path, name, description, is_public, info)

        # DerivedTopic...
        self.__unit = unit                          # string
        self.__derived_data = derived_data          # DerivedData


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        # AbstractTopic...
        jdict['topic'] = self.path
        jdict['name'] = self.name
        jdict['description'] = self.description

        jdict['public'] = self.is_public

        jdict['topic-info'] = self.info

        # DerivedTopic...
        jdict['unit'] = self.unit
        jdict['derived-data'] = self.derived_data

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def unit(self):
        return self.__unit


    @property
    def derived_data(self):
        return self.__derived_data


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DerivedTopic:{path:%s, name:%s, description:%s, is_public:%s, info:%s, " \
               "unit:%s, derived_data:%s}" % \
               (self.path, self.name, self.description, self.is_public, self.info,
                self.unit, self.derived_data)



"""
Created on 2 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/3348460/csv-file-written-with-python-has-blank-lines-between-each-row
"""

import csv
import os
import sys

from scs_core.csv.csv_dict import CSVDict
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class CSVWriter(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, filename=None, append=False, exclude_header=False, header_scan=False, quote_all=False):
        """
        Constructor
        """
        self.__filename = filename
        self.__paths = []

        quoting = csv.QUOTE_ALL if quote_all else csv.QUOTE_MINIMAL

        if self.__filename is None:
            self.__append = append

            self.__file = sys.stdout
            self.__writer = csv.writer(self.__file, quoting=quoting)
        else:
            self.__append = append and os.path.exists(self.__filename)

            if self.__append and not header_scan:
                self.__paths = self.__build_paths()

            self.__file = open(self.__filename, "a" if self.__append else "w", newline='')
            self.__writer = csv.writer(self.__file, quoting=quoting)

        self.__exclude_header = exclude_header
        self.__header_scan = header_scan

        self.__data = []


    # ----------------------------------------------------------------------------------------------------------------

    def write(self, jstr):
        if jstr is None:
            return False

        datum = CSVDict.construct_from_jstr(jstr)

        if datum is None:
            return False

        if self.__header_scan:
            self.__data.append(datum)
            self.__update_paths(datum)

            return True

        if not self.__paths:
            self.__paths = datum.paths()

            # write header...
            if not self.__append and not self.__exclude_header:
                self.__writer.writerow(self.__paths)

        # write row...
        self.__writer.writerow(datum.row(self.__paths))

        if self.filename is None:
            self.__file.flush()

        return True


    def close(self):
        if self.__header_scan:
            # write header...
            self.__writer.writerow(self.__paths)

            # write rows...
            for datum in self.__data:
                self.__writer.writerow(datum.row(self.__paths))

        if self.filename is None:
            return

        self.__file.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __build_paths(self):
        file = sys.stdin if self.__filename is None else open(self.__filename, "r")
        reader = csv.reader(file)

        paths = next(reader)

        file.close()

        return paths


    def __update_paths(self, datum):
        datum_paths = datum.paths()

        appended_paths = []
        for i in range(len(datum_paths)):
            if datum_paths[i] not in self.__paths and not self.__is_sub_path(datum_paths[i], self.__paths):
                self.__paths.insert(i, datum_paths[i])
                appended_paths.append(datum_paths[i])

        if appended_paths:
            for i in reversed(range(len(self.__paths))):
                if self.__is_sub_path(self.__paths[i], appended_paths):
                    self.__paths.pop(i)


    @staticmethod
    def __is_sub_path(candidate, paths):
        for path in paths:
            if candidate == path:
                return False            # exit here because paths are assumed to be unique

            if PathDict.sub_path_includes_path(candidate, path):
                return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def filename(self):
        return self.__filename


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVWriter:{filename:%s, append:%s, exclude_header:%s, header_scan:%s, paths:%s}" % \
               (self.filename, self.__append, self.__exclude_header, self.__header_scan, self.__paths)



"""
Created on 25 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"device": "scs-bgx-401", "topic": "south-coast-science-demo/brighton/loc/1/climate",
"lastSeenTime": "2020-10-23T08:52:20Z", "last_write": "2020-10-23T08:52:20Z",
"message": "{\"val\": {\"hmd\": 68.4, \"tmp\": 19.8, \"bar\": null}, \"rec\": \"2020-10-23T08:52:20Z\",
\"tag\": \"scs-bgx-401\"}"}
"""

import json

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class Byline(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        device = jdict.get('device')
        topic = jdict.get('topic')

        pub = LocalizedDatetime.construct_from_iso8601(jdict.get('lastSeenTime'))
        rec = LocalizedDatetime.construct_from_iso8601(jdict.get('last_write'))

        try:
            jdict.get('message').keys()
            message = json.dumps(jdict.get('message'))      # web API - message is a dict

        except AttributeError:
            message = jdict.get('message')                  # this class - message is a string

        return cls(device, topic, pub, rec, message)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device, topic, pub, rec, message):
        """
        Constructor
        """
        self.__device = device                      # string tag
        self.__topic = topic                        # string path

        self.__pub = pub                            # LocalizedDatetime
        self.__rec = rec                            # LocalizedDatetime

        self.__message = message                    # string


    def __lt__(self, other):
        # device...
        if self.__device < other.__device:
            return True

        if self.__device > other.__device:
            return False

        # topic...
        if self.__topic < other.__topic:
            return True

        if self.__topic > other.__topic:
            return False

        # rec...
        if self.__rec is None:
            return True

        if other.__rec is None:
            return False

        if self.__rec < other.__rec:
            return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self, include_message=True):
        jdict = OrderedDict()

        jdict['device'] = self.device
        jdict['topic'] = self.topic

        jdict['lastSeenTime'] = None if self.pub is None else self.pub.as_iso8601()
        jdict['last_write'] = None if self.rec is None else self.rec.as_iso8601()

        if include_message:
            jdict['message'] = self.message

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        return self.__device


    @property
    def topic(self):
        return self.__topic


    @property
    def pub(self):
        return self.__pub


    @property
    def rec(self):
        return self.__rec


    @property
    def message(self):
        return self.__message


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Byline:{device:%s, topic:%s, pub:%s, rec:%s, message:%s}" %  \
               (self.device, self.topic, self.pub, self.rec, self.message)


# --------------------------------------------------------------------------------------------------------------------

class BylineGroup(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, excluded=None, skeleton=False):
        if not jdict:
            return cls({}) if skeleton else None

        # bylines...
        bylines = []

        for byline_jdict in jdict:
            byline = Byline.construct_from_jdict(byline_jdict)

            if not excluded or not byline.topic.endswith(excluded):
                bylines.append(byline)

        # device_bylines...
        device_bylines = OrderedDict()

        for byline in sorted(bylines):
            if byline.device not in device_bylines:
                device_bylines[byline.device] = []

            device_bylines[byline.device].append(byline)

        return cls(device_bylines)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        self._device_bylines = device_bylines                   # dict of device: Byline


    def __len__(self):
        return len(list(self.bylines))


    # ----------------------------------------------------------------------------------------------------------------

    def latest_topic(self, suffix):
        latest_rec = None
        topic = None

        for byline in self.bylines:
            if byline.topic.endswith(suffix) and (latest_rec is None or byline.rec > latest_rec):
                latest_rec = byline.rec
                topic = byline.topic

        return topic


    def latest_pub(self):
        if not self.bylines:
            return None

        return max([byline.pub for byline in self.bylines if byline.pub is not None])


    def latest_rec(self):
        if not self.bylines:
            return None

        return max([byline.rec for byline in self.bylines if byline.rec is not None])


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return [byline.as_json() for byline in self.bylines]    # matches the structure of the API response


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def bylines(self):
        for bylines in self._device_bylines.values():
            for byline in bylines:
                yield byline


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{device_bylines:%s}" % Str.collection(self._device_bylines)


# --------------------------------------------------------------------------------------------------------------------

class DeviceBylineGroup(BylineGroup):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        super().__init__(device_bylines)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        for device in self._device_bylines.keys():
            return device                                       # return the first device

        return None


# --------------------------------------------------------------------------------------------------------------------

class TopicBylineGroup(BylineGroup):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_bylines):
        """
        Constructor
        """
        super().__init__(device_bylines)


    # ----------------------------------------------------------------------------------------------------------------

    def bylines_for_device(self, device):
        return self._device_bylines[device]                     # may raise KeyError


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def devices(self):
        return list(self._device_bylines.keys())
